/* datetime.c */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "compkey.h"
#include "datetime.h"

/* Contents:
 add_day:      Importing func to add days to dates
 pow10:        Utility to compute powers of 10
 XXX_to_str:   Converting from datetimes (@tz) to string
 str_to_XXX:   Converting strings to datetimes (utc)
 explode_XXX:  Splitting datetimes in their constituents
 implode_XXX:  Constructing datetimes from their constituents
 check_XXX:    Checking validity of datetime constituents
 simparse:     Parse datetime strings according to template
 utc_today:    Calculate current UTC date
 utc_now:      Calculate current UTC time
*/

static uint32 null_add_day(uint32 date, sint32 add) {
 return date;
}
static add_day_fnc add_day=null_add_day;

static int pow10(int n);

static void explode_time1_unchecked(const uint32 *sql_time, int* hh, int* mm, int* ss);
static void implode_time1_unchecked(uint32 *sql_time, int* hh, int* mm, int* ss);


static bool check_time_val(int precision, uint32 hour, uint32 min, uint32 sec, uint32 frac);
static bool check_timezone_val(sint32 zonehour, uint32 zonemin);
static bool check_date_val(uint32 year, uint32 month, uint32 day);

/* add_day:         Importing func to add days to dates */

void set_add_day(add_day_fnc fnc) {
 add_day=fnc;
}

/* pow10:        Utility to compute powers of 10 */

static int pow10(int n) {
 int z=1;
 while(n--) z*=10;
 return z;
}

/* XXX_to_str:     Converting from datetimes (@tz) to string */

bool sql_time_to_str(const uint32 *sql_time, int precision, int timezone, char *pt) {
 int hour, min, sec, frac;

 if(!explode_time(sql_time,precision,timezone,0,&hour,&min,&sec,&frac)) {
     return false;
 }

 sprintf(pt,"%02d:%02d:%02d",hour,min,sec);
 if(precision!=0 && frac!=0) {
     sprintf(pt+8,".%0*d",precision,frac);
 }
 return true;
}

bool sql_date_to_str(const uint32 *now, char *pt) {
 int day, mon, year;

 if(!explode_date(now,&year,&mon,&day)) {
     return false;
 }

 sprintf(pt,"%04d-%02d-%02d",year,mon,day);

 return true;
}

bool sql_timestamp_to_str(const uint32 *sql_timestamp, int precision, int timezone, char *buf) {
 int day, mon, year;
 int hour, min, sec, frac;

 if(!explode_timestamp(sql_timestamp,precision,timezone,&year,&mon,&day,&hour,&min,&sec,&frac)) {
     return false;
 }

 if(precision!=0 && frac!=0) {
     sprintf(buf,"%04d-%02d-%02d %02d:%02d:%02d.%0*d",year,mon,day,hour,min,sec,precision,frac);
 } else {
     sprintf(buf,"%04d-%02d-%02d %02d:%02d:%02d",year,mon,day,hour,min,sec);
 }

 return true;
}

bool sql_interval_to_str(const uint32 *inter, int interval_type, char *buf) {
 int sign;
 int day, mon, year;
 int hour, min, sec, frac;
 if(!explode_interval(inter,interval_type,&sign,&year,&mon,&day,
     &hour,&min,&sec,&frac)) {
     return false;
 }
 switch(interval_type) {
     case T_INTERVAL_YEAR    :
          sprintf(buf,"%d",year); break;
     case T_INTERVAL_MONTH   :
          sprintf(buf,"%d",mon); break;
     case T_INTERVAL_DAY     :
          sprintf(buf,"%d",day); break;
     case T_INTERVAL_HOUR    :
          sprintf(buf,"%d",hour); break;
     case T_INTERVAL_MINUTE  :
          sprintf(buf,"%d",min); break;
     case T_INTERVAL_SECOND  :
          if(frac==0) {
               sprintf(buf,"%d",sec);
          } else {
               sprintf(buf,"%d.%06d",sec,frac);
          }
          break;
     case T_INTERVAL_YEAR_TO_MONTH     :
               sprintf(buf,"%d-%d",year,mon); break;
     case T_INTERVAL_DAY_TO_HOUR       :
               sprintf(buf,"%d %d",day,hour); break;
     case T_INTERVAL_DAY_TO_MINUTE     :
               sprintf(buf,"%d %d:%d",day,hour,min); break;
     case T_INTERVAL_DAY_TO_SECOND     :
               if(frac==0) {
                    sprintf(buf,"%d %d:%d:%d",day,hour,min,sec);
               } else {
                    sprintf(buf,"%d %d:%d:%d.%06d",day,hour,min,sec,frac);
               } break;
     case T_INTERVAL_HOUR_TO_MINUTE    :
               sprintf(buf,"%d:%d",hour,min); break;
     case T_INTERVAL_HOUR_TO_SECOND    :
               if(frac==0) {
                    sprintf(buf,"%d:%d:%d",hour,min,sec);
               } else {
                    sprintf(buf,"%d:%d:%d.%06d",hour,min,sec,frac);
               } break;
     case T_INTERVAL_MINUTE_TO_SECOND  :
               if(frac==0) {
                    sprintf(buf,"%d:%d",min,sec);
               } else {
                    sprintf(buf,"%d:%d.%06d",min,sec,frac);
               } break;
 }
 return true;
}

/* str_to_XXX:      Converting strings to datetimes (utc) */

bool sql_str_to_time(const char *pt, int tz, int *overflow, uint32 *val) {
 int v[6]={0,0,0,0,0,0};
 int le[6];
 int cnt;
 int sign = 1;
 sint32 timezone;
 int has_timezone;
 enum { hour=0, min, sec, frac, zonehour, zonemin};

 if(!simparse(pt,"99:99:99?.8?s99:99",v,le,&cnt)) {
     return 0;
 }

 has_timezone= (cnt==5 || cnt==6);
 if(cnt==5) {
     v[zonemin]=v[frac];
     le[zonemin]=le[frac];
     v[zonehour]=v[zonemin];
     le[zonehour]=le[zonemin];
     v[frac]=0;
     le[frac]=0;
 }

 if(v[frac]!=0) {
     v[frac]*=pow10(6-le[frac]);
 }

 if(!check_time_val(6,v[hour],v[min],v[sec],v[frac])
 || !check_timezone_val(v[zonehour],v[zonemin])) {
     return false;
 }

 if(has_timezone) {
     int sign=1;
     if(v[zonehour]<0) {
          v[zonehour]=-v[zonehour];
          sign=-1;
     }
     timezone=(v[zonehour]*3600+v[zonemin]*60)*sign;
 } else {
     timezone=tz;
 }

 implode_time(val,6,timezone,overflow,v+hour,v+min,v+sec,v+frac);

 return true;
}

bool sql_str_to_timestamp(const char *pt, int tz, uint32 *val) {
 int overflow;
 int v[9]={0,0,0,0,0,0,0,0,0};
 int le[9];
 int cnt;
 int sign = 1;
 sint32 timezone;
 int has_timezone;
 uint32 dat;
 enum { year, month, day, hour, min, sec, frac, zonehour, zonemin};

 if(!simparse(pt,"9999-99-99 99:99:99?.8?s99:99",v,le,&cnt)) {
     return 0;
 }

 has_timezone= (cnt==8 || cnt==9);
 if(cnt==8) {
     v[zonemin]=v[frac];
     le[zonemin]=le[frac];
     v[zonehour]=v[zonemin];
     le[zonehour]=le[zonemin];
     v[frac]=0;
     le[frac]=0;
 }

 if(v[frac]!=0) {
     v[frac]*=pow10(6-le[frac]);
 }

 if(!check_time_val(6,v[hour],v[min],v[sec],v[frac])
 || !check_date_val(v[year],v[month],v[day])
 || !check_timezone_val(v[zonehour],v[zonemin])) {
     return false;
 }

 if(has_timezone) {
     int sign=1;
     if(v[zonehour]<0) {
          v[zonehour]=-v[zonehour];
          sign=-1;
     }
     timezone=(v[zonehour]*3600+v[zonemin]*60)*sign;
 } else {
     timezone=tz;
 }

 implode_time(val,6,timezone,&overflow,v+hour,v+min,v+sec,v+frac);

 implode_date(&dat,v+year,v+month,v+day);
 if(overflow!=0) {
     dat=add_day(dat,overflow);
 }
 Write(uint32,val+2,dat);

 return true;
}

bool sql_str_to_interval(const char *pt, int interval_type ,uint32 *val) {
 int v[7]={0,0,0,0,0,0,0};
 int le[7];
 int cnt;
 int sign=1;
 enum { year, month, day, hour, min, sec, frac};
 const char *full_template="98-98-98 98:98:98?.8";
                         /* -123456789-123456789 */
 const int start_offsets[]={0,3,6,9,12,15};
 const int   end_offsets[]={1,4,7,10,13,19};
 char template[25];
 int start_i, end_i;

 /* Check for initial minus sign */
 if(*pt=='-') {
     pt++;
     sign=-1;
 } else if(*pt=='+') {
     pt++;
     sign=+1;
 }

 /* Which range of values to parse */
 switch(interval_type) {
     case T_INTERVAL_YEAR    : start_i=end_i=year; break;
     case T_INTERVAL_MONTH   : start_i=end_i=month; break;
     case T_INTERVAL_DAY     : start_i=end_i=day; break;
     case T_INTERVAL_HOUR    : start_i=end_i=hour; break;
     case T_INTERVAL_MINUTE  : start_i=end_i=min; break;
     case T_INTERVAL_SECOND  : start_i=end_i=sec; break;
     case T_INTERVAL_YEAR_TO_MONTH     : start_i=year; end_i=month; break;
     case T_INTERVAL_DAY_TO_HOUR       : start_i=day; end_i=hour; break;
     case T_INTERVAL_DAY_TO_MINUTE     : start_i=day; end_i=min; break;
     case T_INTERVAL_DAY_TO_SECOND     : start_i=day; end_i=sec; break;
     case T_INTERVAL_HOUR_TO_MINUTE    : start_i=hour; end_i=min; break;
     case T_INTERVAL_HOUR_TO_SECOND    : start_i=hour; end_i=sec; break;
     case T_INTERVAL_MINUTE_TO_SECOND  : start_i=min; end_i=sec; break;
 }

 strncpy(template,template+start_offsets[start_i],
     end_offsets[end_i]-start_offsets[start_i]);

 if(!simparse(pt,template,v+start_i,le,&cnt)) {
     return 0;
 }

 implode_interval(val,interval_type,&sign,v+year,v+month,v+day,
     v+hour,v+min,v+sec,v+frac);

 return true;
}

bool sql_str_to_date(const char *pt, uint32 *val) {
 int v[3];
 int le[3];
 int cnt;
 enum { year=0, month, day };

 if(!simparse(pt,"9999-99-99",v,le,&cnt)) {
     return false;
 }

 if(!check_date_val(v[year],v[month],v[day])) {
     return false;
 }
 implode_date(val,v+year,v+month,v+day);
 return true;
}

/* explode_XXX:     Splitting datetimes in their constituents */

static unsigned int days[] = {31,28,31,30,31,30,31,31,30,31,30,31 };

static unsigned int days_of_month(int month, int year) {
 if(month!=2) return days[month-1];
 if(year%4==0 && (year%100!=0||year%400==0)) return 29;
 return 28;
}

bool explode_date(const uint32 *pdt, int* YY, int* MM, int* DD) {
 uint32 dt=Read(uint32,pdt);
 int dd,mm,yy;
 dd = (int)(dt % 256);
 mm = (int)((dt >> 8) % 256);
 yy = (int)(dt >> 16);

 if(!check_date_val(yy,mm,dd)) {
     return false;
 }

 *DD=dd; *MM=mm; *YY=yy;
 return true;
}

static void explode_time1_unchecked(const uint32 *sql_time, int* hh, int* mm, int* ss) {
 sint32 t;
 uint32 hour, min, sec;
 t=Read(sint32,sql_time);
 hour=t/3600;
 min=(t - hour*3600)/60;
 sec=t%60;

 *hh=hour; *mm=min; *ss=sec;
}

bool explode_time(const uint32 *sql_time, int precision, int timezone, int *overflow,
     int* hh, int* mm, int* ss, int *fr) {
 sint32 t;
 uint32 hour, min, sec, frac=0;
 static const int daysecs=24*60*60;
 int of=0;

 t=Read(sint32,sql_time);
 t-=timezone;
 if(t>daysecs) {
     t-=daysecs;
     of=+1;
 } else if(t<0) {
     t+=daysecs;
     of=-1;
 }

 hour=t/3600;
 min=(t - hour*3600)/60;
 sec=t%60;
 if(precision!=0) {
      frac=Read(uint32,sql_time+1);
 }      

 if(!check_time_val(precision,hour,min,sec,frac)) {
     return false;
 }

 *hh=hour; *mm=min; *ss=sec;
 if(fr!=0) {
     *fr=frac;
 }
 if(overflow!=0) {
     *overflow=of;
 }

 return true;
}

bool explode_timestamp(const uint32 *now, int precision, int tz,
     int* YY, int* MM, int* DD,
     int* hh, int* mm, int* ss, int *frac) {
 int overflow;
 uint32 dt;

 if(!explode_time(now,precision,tz,&overflow,hh,mm,ss,frac)) {
     return false;
 }
 if(precision==0) {
     dt=Read(uint32,now+1);
 } else {
     dt=Read(uint32,now+2);
 }     
 if(overflow!=0) {
     dt=add_day(dt,overflow);
 }
 if(!explode_date(&dt,YY,MM,DD)) {
     return false;
 }
 return true;
}

static void explode_sign(const uint32 *inp, int *sign, int *val) {
 sint32 v=Read(sint32,inp);
 if(v>=0) {
     *sign=+1;
     *val=v;
 } else {
     *sign=-1;
     *val=-v;
 }
}

bool explode_interval(const uint32 *v, int interval_type,
     int *sign, int* YY, int* MM, int* DD,
     int* hh, int* mm, int* ss, int *frac) {
 uint32 tm;

 tm=*sign=*YY=*MM=*DD=*hh=*mm=*ss=*frac=0;
 switch(interval_type) {
     case T_INTERVAL_YEAR    : explode_sign(v,sign,YY); break;
     case T_INTERVAL_MONTH   : explode_sign(v,sign,MM); break;
     case T_INTERVAL_DAY     : explode_sign(v,sign,DD); break;
     case T_INTERVAL_HOUR    : explode_sign(v,sign,hh); break;
     case T_INTERVAL_MINUTE  : explode_sign(v,sign,mm); break;
     case T_INTERVAL_SECOND  :
          explode_sign(v,sign,ss);
          *frac=Read(uint32,v+1);
          break;
     case T_INTERVAL_YEAR_TO_MONTH     :
          explode_sign(v,sign,YY);
          *MM=Read(uint32,v+1);
          break;
     case T_INTERVAL_DAY_TO_HOUR       :
     case T_INTERVAL_DAY_TO_MINUTE     :
          explode_sign(v,sign,DD);
          tm=Read(uint32,v+1);
          break;
     case T_INTERVAL_DAY_TO_SECOND     :
          explode_sign(v,sign,DD);
          tm=Read(uint32,v+1);
          *frac=Read(uint32,v+2);
          break;
     case T_INTERVAL_HOUR_TO_MINUTE    :
          explode_sign(v,sign,&tm);
          break;
     case T_INTERVAL_HOUR_TO_SECOND    :
     case T_INTERVAL_MINUTE_TO_SECOND  :
          explode_sign(v,sign,&tm);
          *frac=Read(uint32,v+1);
          break;
 }
 explode_time1_unchecked(&tm,hh,mm,ss);
 return true;
}

/* implode_XXX:     Constructing datetimes from their constituents */

bool implode_date(uint32 *d, int* YY, int* MM, int* DD) {
 uint32 dat=(*YY << 16) + (*MM << 8) + *DD;
 Write(uint32,d,dat);
 return true;
}

static void implode_time1_unchecked(uint32 *sql_time, int* hh, int* mm, int* ss) {
 uint32 tim;
 tim=*hh*3600+*mm*60+*ss+timezone;
 Write(uint32,sql_time,tim);
}

bool implode_time(uint32 *sql_time, int precision, int timezone, int *overflow,
     int* hh, int* mm, int* ss, int *fr) {
 sint32 tim;
 static const int daysecs=24*60*60;

 if(!check_time_val(precision,*hh,*mm,*ss,*fr)) {
     return false;
 }
 tim=*hh*3600+*mm*60+*ss+timezone;
 *overflow=0;
 if(tim>=daysecs) {
     *overflow=+1;
     tim-=daysecs;
 } else if(tim<0) {
     *overflow=-1;
     tim+=daysecs;
 }
 Write(uint32,sql_time,tim);
 if(precision!=0) {
     Write(uint32,sql_time+1,*fr);
 }     

 return true;
}

bool implode_timestamp(uint32 *now, int precision, int tz,
     int* YY, int* MM, int* DD,
     int* hh, int* mm, int* ss, int *frac) {
 uint32 dat;
 int overflow;

 if(!implode_time(now,precision,tz,&overflow,hh,mm,ss,frac)) {
     return false;
 }
 if(!implode_date(&dat,YY,MM,DD)) {
     return false;
 }
 if(overflow!=0) {
     dat=add_day(dat,overflow);
 }
 if(precision==0) {
     Write(uint32,now+1,dat);
 } else {
     Write(uint32,now+2,dat);
 }
 return true;
}

bool implode_interval(uint32 *v, int interval_type,
     int *sign, int* YY, int* MM, int* DD,
     int* hh, int* mm, int* ss, int *frac) {
 uint32 tm;

 implode_time1_unchecked(&tm,hh,mm,ss);
 switch(interval_type) {
     case T_INTERVAL_YEAR    : Write(sint32,v,*YY*(*sign)); break;
     case T_INTERVAL_MONTH   : Write(sint32,v,*MM*(*sign)); break;
     case T_INTERVAL_DAY     : Write(sint32,v,*DD*(*sign)); break;
     case T_INTERVAL_HOUR    : Write(sint32,v,*hh*(*sign)); break;
     case T_INTERVAL_MINUTE  : Write(sint32,v,*mm*(*sign)); break;
     case T_INTERVAL_SECOND  :
          Write(sint32,v,*ss*(*sign));
          Write(uint32,v+1,*frac);
          break;
     case T_INTERVAL_YEAR_TO_MONTH     :
          Write(sint32,v,*YY*(*sign));
          Write(uint32,v+1,*MM);
          break;
     case T_INTERVAL_DAY_TO_HOUR       :
     case T_INTERVAL_DAY_TO_MINUTE     :
          Write(sint32,v,*DD*(*sign));
          Write(uint32,v+1,tm);
          break;
     case T_INTERVAL_DAY_TO_SECOND     :
          Write(sint32,v,*DD*(*sign));
          Write(uint32,v+1,tm);
          Write(uint32,v+2,*frac);
          break;
     case T_INTERVAL_HOUR_TO_MINUTE    :
          Write(sint32,v,tm*(*sign));
          break;
     case T_INTERVAL_HOUR_TO_SECOND    :
     case T_INTERVAL_MINUTE_TO_SECOND  :
          Write(sint32,v,tm*(*sign));
          Write(uint32,v+1,*frac);
          break;
 }
 return true;
}

/* check_XXX:  Checking validity of datetime constituents */

static bool check_time_val(int precision, uint32 hour, uint32 min, uint32 sec, uint32 frac) {
 /* Note: sec can be up to 61 (leap seconds) */
 if(hour>23 || min>59 || sec>61 || frac>=(uint32)pow10(precision)) {
     return false;
 }
 return true;
}

static bool check_timezone_val(sint32 zonehour, uint32 zonemin) {
 /* -12:59 to +13:00 */
 if(zonehour<-12 || zonehour>13) {
     return false;
 }
 if(zonehour==13 && zonemin>0) {
     return false;
 }
 if(zonemin>59) {
     return false;
 }
 return true;
}

static bool check_date_val(uint32 yy, uint32 mm, uint32 dd) {
 if(mm>12 || dd>days_of_month(mm,yy)) {
     return false;
 }
 return true;
}

/* simparse:   Parse datetime strings according to template */
/*   Character      Meaning
     ---------      -----------------------------------------------------
     9              A single digit
     8              Zero or more digits
     s              Sign (+ or -) incorporated in next number, if any
     ?              Input could end here or continue
     any other      The character itself
*/

int simparse(const char *input, const char *stempl, int *output, int *lengths, int *ocount) {
 int optional_flag=0;
 int icurr=0;
 char buffer[30];
 char *p=buffer;
 const char *templ=stempl;

 while(*input!=0) {
  switch(*templ) {
     case 0: return 0;
     case '?': optional_flag=1; templ++; break;
     case '9': templ++; /*fallthrough*/
     case '8':
          optional_flag=0;
          if(isdigit(*input)) {
               *p++=*input++;
          } else if(*templ=='8') {
               templ++;
          } else {
               return 0;           /* expected digit */
          }
          break;
     default:
          optional_flag=0;
          if(*templ==*input || (*templ=='s' && (*input=='+' || *input=='-'))) {
               if(p!=buffer) {
                    *p=0;
                    lengths[icurr]=strlen(buffer);
                    output[icurr++]=atoi(buffer);
                    p=buffer;
               }
               if(*templ=='s') {
                    *p++=*input;
               }
               input++;
               templ++;
          } else {
               return 0;           /* expected specific char */
          }
  }
 }
 if(p!=buffer) {
     *p=0;
     lengths[icurr]=strlen(buffer);
     output[icurr++]=atoi(buffer);
 }
 if(optional_flag || *templ==0 || *templ=='?' || *templ=='8') {
     *ocount=icurr;
     return 1;
 } else {
     *ocount=0;
     return 0;
 }
}


/* utc_today:  Calculate current UTC date */

uint32 utc_today() {
 uint32 val;
 int year,month,day;

 time_t now;
 struct tm *tmnow;

 now = time(0);
 tmnow = gmtime(&now);
 year = tmnow->tm_year + 1900;
 month = tmnow->tm_mon+1;
 day = tmnow->tm_mday;
 implode_date(&val,&year,&month,&day);
 return val;
}

/* utc_now:    Calculate current UTC time */

uint32 utc_now() {
 uint32 val;
 int hour,min,sec,fr=0,ov;

 time_t now;
 struct tm *tmnow;

 now = time(0);
 tmnow = gmtime(&now);
 hour = tmnow->tm_hour;
 min = tmnow->tm_min;
 sec = tmnow->tm_sec;
 implode_time(&val,0,0,&ov,&hour,&min,&sec,&fr);
 return val;
}
