/* http.c */

#include <fcntl.h>
#ifdef NT
#include <io.h>
#else
#include <unistd.h>
#endif
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "err_sys.h"
#include "openexec.h"
#include "spawn.h"
#include "ipc.h"
#include "stricmp.h"
#include "datetime.h"
#include "config.h"
#include "http.h"

#if defined(UNIX)
# define OPENMODE_R O_RDONLY
#elif defined(NT)
#include <io.h>
# define OPENMODE_R (_O_RDONLY | _O_BINARY)
#endif

#if defined(BSD)
# define BSD_SENDMSG
# undef _T
#endif

#define CGI_BUFFER_SZ (4*1024)

enum {
 MAXSESSIONS=200
};

typedef enum {
 req_parse200ok, req_parse400bad, req_parse414long, req_parse404noses
} req_parse_status;

/* Default values for configuration parameters */
static handler_s default_handlers[]={
{"/",     file_handler}
};
static auth_outcome default_authenticate(const char *u, const char *p, const char* host, sint16 *i) {
 *i=0;
 return auth_ok;
}

static void default_shutdown() {
 exit(1);
}

static bool critical_nop() {
 return true;
}

static bool default_email_to_user(char username[], sint16 *puid, const char email[]) {
 return false;
}

/* Configuration parameters */
static uint32 my_IP_address;
static char my_hostname[129];
static int http_port;
static bool is_secure;
static handler_s *handlers=default_handlers;
static int handlers_nb=sizeof(default_handlers)/sizeof(default_handlers[0]);
static mime_type_handler_s *mime_type_handlers=0;
static int mime_type_handlers_nb=0;
static auth_func http_authenticate=default_authenticate;
static email_to_user_func email_to_user=default_email_to_user;
static int sslauthentication=AUTH_REQ_PASSWORD;
static shutdown_func down=default_shutdown;
static critical_section_func critical_enter=critical_nop, critical_leave=critical_nop;
static char config_httplogfile[TOKSIZE]="http.log";
static char config_httprootdir[TOKSIZE]="./";
static char config_httpshareddir[TOKSIZE]="";
static char config_execext[TOKSIZE]="";
static char config_binpath[TOKSIZE]="./";
static char config_mimetypesfile[TOKSIZE]="";
static int config_httpsessiontimeout=15;
static char server_token[TOKSIZE]="Ovrimos-Http-Server/2.0";
static char **additional_env_vars;
static int additional_env_vars_nb=0;

/* HTTP sessions */
static void session_data_init(session_data *);
static void session_data_reinit(session_data *);
static realm_data *session_data_get_realm_data(session_data *s, const char *realm);
static void session_data_set_realm_data(session_data *s, const char *realm, realm_data *data);
static void session_reap();
static uint32 session_new_ID(const char *user, struct sockaddr *peer);
static bool session_valid(uint32 sessionID, struct sockaddr *peer);
static bool session_set_user(uint32 sessionID, const char *user);
static void session_grab(uint32 sessionID, int incr);

typedef struct session_cell {
 uint32 sessionID;
 time_t created;
 time_t accessed;
 int in_use;
 char init_user[50];
 char user[50];
 struct sockaddr peer_addr;
 session_data data;
 struct session_cell *next;
 struct session_cell *prev;
} session_cell;

static int session_num;
static session_cell *sessions;
static time_t last_aging;

static session_cell *find_session_cell(uint32 sessionID, struct sockaddr *peer, bool boost);
static session_cell *make_session_cell(uint32 sessionID, const char *user, struct sockaddr *peer);

static session_cell _sessions;

static bool http_status_code(PS_SOCKET so, http_header *phd, hdl_t ret);
static void http_404_not_found(PS_SOCKET so, http_header *phd);
static void http_404_no_session(PS_SOCKET so, http_header *phd);
static req_parse_status http_parse_request(const char *line, http_header *phd);
static req_parse_status http_header_fill(http_header *,PS_SOCKET so);
static req_parse_status http_read_form_data(http_header *,PS_SOCKET so);
static void http_header_init(http_header *, struct sockaddr *peer);
static void http_header_deinit(http_header *p);
static void http_header_copy(http_header *dest, http_header *src);
static void http_header_form_reinit(http_header *p);
static req_parse_status http_header_update(http_header*, const char *header_line);
static void http_log_request(http_header *, PS_SOCKET);

/* Status codes are returned (and recorded) by http_request and 
   handler_read_form_data */
static void form_data_init(form_data *);
static void form_data_abandon_files(form_data *p);
static void form_data_deinit(form_data *);
static void form_data_copy(form_data *dest, form_data *src);
static bool form_data_multipart_read(form_data *, const char *boundary, PS_SOCKET so);
static bool form_data_url_decode(form_data *, const char *line, bool isindex_allowed);
static bool form_data_url_encode(form_data *, char *line);
static void form_data_add(form_data *, const char *name, const char *data, bool is_file);

static void format_time(char *buffer, time_t ti);
static void free_strduped_env(char **env);
static bool is_nph_cgi(const char *cmdline);
static int indirect_strcmp(const void *p1, const void *p2);
static int url_decode(const char *urlenc, char *plain);
static int url_encode(const char *plain, char *urlenc);
static http_req_t http_which_method(const char *line);
static void http_map_ext_to_mime_type(const char *ext, char *mime_type);
static bool same_host(struct sockaddr *p1, struct sockaddr *p2);
static random_seed(uint32 s);
static uint32 random_next();

static bool filetype_is_executable(const char *ext);
#ifdef NT
static bool filetype_open_command_string(const char *ext, const char *fname,
     char *cmdbuffer);
#endif
static bool mimetype_run_command_string(const char *mimetype, const char *fname,
     char *cmdbuffer);
/* HTTP sessions */

static void so_redirect_to_session(PS_SOCKET so, http_header *phd, uint32 sessionID);

static void base64encode(const char *plain, char* base64, int plainlen);
static void base64decode(const char *base64, char* plain, int base64len);
static void base64decode_quadruplet(const uint8 *base64, uint8 *plain);
static void base64encode_triplet(const uint8 *plain, uint8 *base64);
static void base64decode_digit(const uint8 *base64, uint8 *plain);
static void base64encode_digit(const uint8 *plain, uint8 *base64);

/* Begin code */

/* Initialization/Configuration */
void http_server_init(char *hostname, uint32 hostaddr, int port, bool is_sec) {
 strcpy(my_hostname,hostname);
 my_IP_address=hostaddr;
 http_port=port;
 is_secure=is_sec;
 session_initialize();
 random_seed(time(0));
}     

void http_server_set_critical_section_primitives(
     critical_section_func enter,
     critical_section_func leave) {
 critical_enter=enter;
 critical_leave=leave;
}     

void http_server_set_handlers(handler_s *h, int h_nb) {
 handlers=h;
 handlers_nb=h_nb;
}

void http_server_set_mime_type_handlers(mime_type_handler_s *m, int m_nb) {
 mime_type_handlers=m;
 mime_type_handlers_nb=m_nb;     
}

void http_server_set_auth_func(auth_func auth) {
 http_authenticate=auth;
}

void http_server_set_sslauthentication(int i) {
 sslauthentication=i;
}

void http_server_set_shutdown_func(shutdown_func s) {
 down=s;
}

void http_server_set_email_to_user_func(email_to_user_func email2user) {
 email_to_user=email2user;
}

void http_server_set_httplogfile(const char *s) {
 strcpy(config_httplogfile,s);
}

void http_server_set_httprootdir(const char *s) {
 strcpy(config_httprootdir,s);
}

void http_server_set_httpshareddir(const char *s) {
 strcpy(config_httpshareddir,s);
}

void http_server_set_execext(const char *s) {
 strcpy(config_execext,s);
}

void http_server_set_binpath(const char *s) {
 strcpy(config_binpath,s);
}

void http_server_set_mimetypesfile(const char *s) {
 strcpy(config_mimetypesfile,s);
}

void http_server_set_httpsessiontimeout(int i) {
 config_httpsessiontimeout=i;
}

void http_server_add_to_product_token(const char *s) {
 strcat(server_token," ");
 strcat(server_token,s);
}

void http_server_additional_env_vars(char **pp, int nb) {
 additional_env_vars=pp;
 additional_env_vars_nb=nb;
}

const char *http_server_product_token() {
 return server_token;
}
  
/* Service routine */
void http_request(PS_SOCKET so, struct sockaddr *peer, char *buffer) {
 http_header hd;
 int i;
 hdl_t ret = -1;
int rso;
struct linger li;

li.l_onoff = 1;
li.l_linger = 300;

 {
 int ret;
 int no = 1;

	rso = PS_getPureSocket(so);
/*	ret = setsockopt(rso,IPPROTO_TCP,TCP_NODELAY,(char *)&no,sizeof(int));*/
    ret = setsockopt(rso,SOL_SOCKET,SO_KEEPALIVE,(char *)&no,sizeof(int));
	ret = setsockopt(rso,SOL_SOCKET,SO_LINGER,(char *)&li,sizeof(li));

 }
 session_reap();

 http_header_init(&hd,peer);

 switch(http_header_fill(&hd,so)) {
     case req_parse200ok: break;
     case req_parse400bad:
          PS_printf(so,"HTTP/1.0 400 Bad request\r\n\r\n");
          strcpy(hd.status_code,"400");
          goto end;
     case req_parse404noses:
       		 http_404_no_session(so,&hd);
          goto end;     
     case req_parse414long:
          PS_printf(so,"HTTP/1.0 414 Request-URI too long\r\n\r\n");
          strcpy(hd.status_code,"414");
          goto end;
 }

 /* While in a session, requesting a URL without session info results
    in a redirection containing the session info. Thus, all pages in a
    session contain the session info. If, however, the refering session
    is invalid (or has expired) the request is in error. */
 if(hd.sessionID==0 && hd.referer_sessionID!=0) {
     if(!session_valid(hd.referer_sessionID,&hd.peer_addr)) {
          http_404_no_session(so,&hd);
     } else {
          so_redirect_to_session(so,&hd,hd.referer_sessionID);
     }
     goto end;
 }

 for(i=0; i<handlers_nb; i++) {
     if(memicmp(hd.request_URI,handlers[i].prefix,strlen(handlers[i].prefix))!=0) {
          continue;
     }
     if(hd.sessionID!=0) {
          session_grab(hd.sessionID,+1);
     }
     ret=handlers[i].func(so,handlers[i].prefix,buffer,&hd);
     if(hd.sessionID!=0) {
          session_grab(hd.sessionID,-1);
     }
     /* If hdl_sep, try next handler */
     if(ret==hdl_redirect) {
          i=-1; /* So that it will start from 0 */
          http_header_form_reinit(&hd);
     }
     if(http_status_code(so,&hd,ret)) {
          goto end;
     }
 }

 http_404_not_found(so,&hd);

end:
 if(ret!=hdl_defer) {
     if(ret!=hdl_defer_deinit) {
          http_log_request(&hd,so);
     }
	 
     PS_closesocket(so);
     PS_release(so);
     http_header_deinit(&hd);
 }
 if(ret==hdl_down) {
     down();
 }
}

bool http_demands_wap(http_header *phd) {
 /* UP.Phone emulator pretends to accept text/html, but chokes on it... */
 if(memicmp(phd->user_agent,"UP",2)==0) {
     return true;
 }
 /* Opera accepts WML, but HTML is more appropriate */
 if(memicmp(phd->user_agent,"Opera",5)==0) {
     return false;
 }
 if(strstr(phd->accept,"text/vnd.wap.wml")!=0) {
     return true;
 } else {
     return false;
 }
}

static void http_404_not_found(PS_SOCKET so, http_header *phd) {
 PS_printf(so,"HTTP/1.0 404 Not found\r\n");
 strcpy(phd->status_code,"404");
 if(http_demands_wap(phd)) {
     PS_printf(so,"Content-type: text/vnd.wap.wml\r\n\r\n");
     PS_printf(so,
          "<?xml version=\"1.0\"?>\n"
          "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
          "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
          "<wml>\n"
          "<card id=\"main\" title=\"Not found\">\n");
     PS_printf(so,"The requested object %s was not found<br/><a href=\"/\">Home</a>",
          phd->request_URI);
     PS_printf(so,
          "</p>\n"
          "</card>\n"
          "</wml>\n");
 } else {
     PS_printf(so,"Content-type: text/html\r\n\r\n");
     PS_printf(so,"<html><body bgcolor=\"#eeffff\">"
          "<h3>The requested object %s was not found<br><a href=\"/\">Home</a></h3></body></html>",
          phd->request_URI);
     }
}

static void http_404_no_session(PS_SOCKET so, http_header *phd) {
 char buff[1024];

	PS_printf(so,"HTTP/1.0 404 Not found\r\n");
    
	
 strcpy(phd->status_code,"404");
 if(http_demands_wap(phd)) {
     PS_printf(so,"Content-type: text/vnd.wap.wml\r\n\r\n");
     PS_printf(so,
          "<?xml version=\"1.0\"?>\n"
          "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
          "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
          "<wml>\n"
          "<card id=\"main\" title=\"Not found\">\n");
     PS_printf(so,"The session specified was not found<br/><a href=\"/\">Home</a>");
     PS_printf(so,
          "</p>\n"
          "</card>\n"
          "</wml>\n");
 } else {
	 char buf[100];
	 time_t xx;
	 int len;
	
	 sprintf(buff,"<html><head><TITLE>Session Missing!!!</TITLE>"
		 "<META HTTP-EQUIV=\"REFRESH\" CONTENT=\"15; URL=/\"></head>"
		 "<body bgcolor=\"#eeffff\">"
		 "The session you requested is missing. You will be redirected to the server's "
		"<A HREF=\"/\">main page</A> in 15 seconds."
		 "</body></html>  ");
     len = strlen(buff);
	
	 time(&xx);
    format_rfc1123_time(buf,xx);
	PS_printf(so,"%s\r\n",buf);
	 PS_printf(so,"Server: %s\r\n",server_token);
	 PS_printf(so,"Content-Length: %d\r\n",len);
	 PS_printf(so,"Content-type: text/html\r\nPragma: no-cache\r\n\r\n");
	 PS_printf(so,buff);
		 
	
     }
}

void http_redirect(PS_SOCKET so, http_header *phd, const char *fmt, ...) {
 char buffer[TOKSIZE];
 int len;
 va_list marker;

 va_start(marker,fmt);
 len=vsprintf(buffer,fmt,marker);
 va_end(marker);

 if(http_demands_wap(phd)) {
     PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n"
          "Location: %s\r\n", buffer);
     PS_printf(so,"Content-type: text/vnd.wap.wml\r\n\r\n");
     PS_printf(so,
          "<?xml version=\"1.0\"?>\n"
          "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
          "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
          "<wml>\n"
          "<card id=\"main\" title=\"Moved temporarily\">\n");
     PS_printf(so,"The requested object will be found at location %s",
          buffer);
     PS_printf(so,
          "</p>\n"
          "</card>\n"
          "</wml>\n");
 } else {
     PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n"
          "Content-type: text/html\r\n"
          "Location: %s\r\n\r\n", buffer);
     PS_printf(so,"<html><body>The requested object will be "
          "found at location %s</body></html>",
          buffer);
 }
 strcpy(phd->status_code,"302");
/* PS_closesocket(so); Why was it here??? */
}

static void so_redirect_to_session(PS_SOCKET so, http_header *phd, uint32 sessionID) {
 char s_base[512];
 format_base_URI(s_base,phd);

 if(http_demands_wap(phd)) {
     PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n"
          "Location: %s/session%X%s\r\n",
          s_base,sessionID,phd->request_URI);
     PS_printf(so,"Content-type: text/vnd.wap.wml\r\n\r\n");
     PS_printf(so,
          "<?xml version=\"1.0\"?>\n"
          "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
          "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
          "<wml>\n"
          "<card id=\"main\" title=\"Moved temporarily\">\n");
     PS_printf(so,"The requested object will be found at "
          "location %s/session%X%s",
          s_base,sessionID,phd->request_URI);
     PS_printf(so,
          "</p>\n"
          "</card>\n"
          "</wml>\n");
 } else {
     PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n"
          "Content-type: text/html\r\n"
          "Location: %s/session%X%s\r\n\r\n",
          s_base,sessionID,phd->request_URI);
     PS_printf(so,"<html><body>The requested object will be "
          "found at location %s/session%X%s</body></html>",
          s_base,sessionID,phd->request_URI);
 }
 strcpy(phd->status_code,"302");
}

static bool http_status_code(PS_SOCKET so, http_header *phd, hdl_t ret) {
 if(ret==hdl_ok) {
     return true;
 }
 if(ret==hdl_newses) {
     uint32 sessionID=session_new_ID(phd->user,&phd->peer_addr);
     so_redirect_to_session(so,phd,sessionID);
     return true;
 }
 if(ret==hdl_error) {
     PS_printf(so,"HTTP/1.0 500 Internal server error\r\n");
     strcpy(phd->status_code,"500");
     return true;
 }
 if(ret==hdl_meth) {
     PS_printf(so,"HTTP/1.0 405 Method Not Allowed\r\n");
     strcpy(phd->status_code,"405");
     return true;
 }
 if(ret==hdl_down) {
     return true;
 }
 if(ret==hdl_defer) {
     return true;
 }
 if(ret==hdl_defer_deinit) {
     form_data_abandon_files(&phd->form);
     return true;
 }
 if(ret==hdl_auth) {
     if(sslauthentication & AUTH_REQ_PASSWORD) {
          const char *mesg;
          PS_printf(so,"HTTP/1.0 401 Unauthorized\r\n");
          strcpy(phd->status_code,"401");
          PS_printf(so,"WWW-Authenticate: basic realm=\"/\"\r\n");
          switch(phd->auth) {
               case auth_none:
                    mesg="You must login as a valid user";
                    break;
               case auth_failed_passwd:
                    mesg="You must login correctly as a valid user";
                    break;
               case auth_unsup:
                    mesg="This authorization scheme is not supported";
                    break;
               case auth_ok:
                    mesg="You must login as <i>admin</i>";
                    break;
          }
          if(http_demands_wap(phd)) {
               PS_printf(so,
                    "<?xml version=\"1.0\"?>\n"
                    "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
                    "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
                    "<wml>\n"
                    "<card id=\"main\" title=\"Unauthorized\">\n"
                    "<p>\n"
                    "<b>%s</b><br/>\n",
                    mesg);
               PS_printf(so,
                    "</p>\n"
                    "</card>\n"
                    "</wml>\n");
          } else {
               PS_printf(so,"Content-type: text/html\r\n\r\n");
               PS_printf(so,"<html><body bgcolor=\"#eeffff\"><h3>");
               PS_printf(so,"%s</h3></body></html>",mesg);
          }
     } else {
          strcpy(phd->status_code,"401");
          if(http_demands_wap(phd)) {
               PS_printf(so,
                    "HTTP/1.0 401 Unauthorized\r\n"
                    "Content-type: text/vnd.wap.wml\r\nPragma: no-cache\r\n\r\n"
                    "<?xml version=\"1.0\"?>\n"
                    "<!DOCTYPE wml PUBLIC \"-//WAPFORUM//DTD WML 1.1//EN\"\n"
                    "    \"http://www.wapforum.org/DTD/wml_1.1.xml\">\n"
                    "<wml>\n"
                    "<card id=\"main\" title=\"Unauthorized\">\n"
                    "<p>\n"
                    "<b>Your credentials do not prove you are a valid user</b>\n"
                    "</p>\n"
                    "</card>\n"
                    "</wml>\n");
          } else {
               PS_printf(so,"HTTP/1.0 401 Unauthorized\r\n"
                    "Content-type: text/html\r\nPragma: no-cache\r\n\r\n"
                    "<html><head><title>Unauthorized</title></head>\n"
                    "<body><h1>Unauthorized</h1>Your credentials do not prove you are a valid "
                    "user. </body></html>\n");
          }               
     }
     return true;
 }
 return false;
}

static req_parse_status http_header_fill(http_header *phd, PS_SOCKET so) {
 char line[TOKSIZE];
 char email[MAX_EMAIL_SZ] = "";
 req_parse_status main_ret;
 req_parse_status ret;

 if(!PS_recv_line(so,line,sizeof(line)-1)) {
     return req_parse414long;
 }
 phd->orig_meth=phd->meth=http_which_method(line);
 main_ret=http_parse_request(line,phd);
 if(main_ret!=req_parse200ok && main_ret!=req_parse404noses) {
     return main_ret;
 }

 if(phd->full_messages) {
     /* skip header */
     do {
          if(!PS_recv_line(so,line,sizeof(line)-1)) {
               return req_parse400bad;
          }
          ret=http_header_update(phd,line);
          if(ret!=req_parse200ok) {
               return ret;
          }
     } while(*line!=0);
 }

 /* If peer sent certificate, check it, regardless of whether we demanded
    it or not */
 if(PS_getPeerEMail(so, email)!=0) {
     char username[21] = "";
     /* If it doesn't correspond to a user, it's an error */
     if(!email_to_user(username, &phd->uid, email)) {
          phd->auth=auth_failed_cert;
     } else if(phd->auth==auth_ok) {
          /* If username was provided, the two must match */
          if(stricmp(username, phd->user)!=0) {
               phd->auth=auth_failed_cert;
          }
     } else {
          if(!(sslauthentication & AUTH_REQ_PASSWORD)) {
               /* If password required and not given, cert. is not enough */
               /* If only the certificate was given, it determines the username */
               strcpy(phd->user,username);
               phd->auth=auth_ok;
          }
     }
 } else if(sslauthentication & AUTH_REQ_CERT) {
     /* Peer did not send a certificate, but we demand one */
     phd->auth=auth_failed_cert;
 }

 if(phd->sessionID!=0 && phd->auth==auth_ok) {
     session_set_user(phd->sessionID,phd->user);
 }
 
 return main_ret;
}

bool handler_read_form_data(http_header *phd, PS_SOCKET so) {
 switch(http_read_form_data(phd,so)) {
     case req_parse200ok:
          return true;
     case req_parse400bad:
          PS_printf(so,"HTTP/1.0 400 Bad request\r\n");
          strcpy(phd->status_code,"400");
          return false;
     case req_parse414long:
          PS_printf(so,"HTTP/1.0 414 Request-URI too long\r\n");
          strcpy(phd->status_code,"414");
          return false;
 }
 return true;
}

static req_parse_status http_read_form_data(http_header *phd, PS_SOCKET so) {
 char line[TOKSIZE];
 const char *url_enc="application/x-www-form-urlencoded";
 const char *form_data="multipart/form-data";
 char *qmark;

 /* Ensure idempotence when called a second time */
 if(!form_data_empty(&phd->form)) {
     return req_parse200ok;
 }

 if((phd->meth==http_get || phd->meth==http_head)
 && (qmark=strchr(phd->request_URI,'?'))!=0) {
     if(!form_data_url_decode(&phd->form, qmark+1, true)) {
          return req_parse400bad;
     }
     *qmark=0;
 } else if((phd->meth==http_post || phd->meth==http_head)
 && memicmp(phd->content_type,url_enc,strlen(url_enc))==0) {
     if(!PS_recv_all(so,line,phd->content_length)) {
          return req_parse400bad;
     }
     line[phd->content_length]=0;
     if(!form_data_url_decode(&phd->form, line, false)) {
          return req_parse400bad;
     }
 } else if((phd->meth==http_post || phd->meth==http_head)
 && memicmp(phd->content_type,form_data,strlen(form_data))==0) {
     char boundary[100]; /* 70 suffices per RFC 1521 */
     const char *b_sep="; boundary=";
     const char *p=phd->content_type+strlen(form_data);
     if(memicmp(p,b_sep,strlen(b_sep))==0) {
          p+=strlen(b_sep);
          if(*p=='"') {
               strcpy(boundary,p+1);
               boundary[strlen(boundary)-1]=0;
          } else {
               strcpy(boundary,p);
          }
          if(!form_data_multipart_read(&phd->form, boundary,so)) {
               return req_parse400bad;
          }
          phd->content_length=0;
     }
 }

 return req_parse200ok;
}

static void http_header_init(http_header *p, struct sockaddr *peer) {
 *p->host=0;
 *p->referer=0;
 p->content_length=0;
 *p->content_type=0;
 p->auth=auth_none;
 p->uid=-1;
 *p->user=0;
 *p->passwd=0;
 *p->path_info=0;
 *p->user_agent=0;
 *p->cookie=0;
 *p->accept=0;
 p->sessionID=0;
 p->referer_sessionID=0;
 *p->request_URI=0;
 *p->orig_request_URI=0;
 memcpy(&p->peer_addr,peer,sizeof(struct sockaddr));
 form_data_init(&p->form);
 form_data_init(&p->other_headers);
}

static void http_header_deinit(http_header *p) {
 form_data_deinit(&p->form);
 form_data_deinit(&p->other_headers);
}

static void http_header_copy(http_header *dest, http_header *src) {
 memcpy(dest,src,sizeof(http_header));
 form_data_copy(&dest->form,&src->form);
 form_data_copy(&dest->other_headers,&src->other_headers);
}

static void http_header_form_reinit(http_header *p) {
 form_data_deinit(&p->form);
 form_data_deinit(&p->other_headers);
 form_data_init(&p->form);
 form_data_init(&p->other_headers);
}


static void http_log_request(http_header *p, PS_SOCKET so) {
 FILE *logfile=fopen(config_httplogfile,"a");
 char peer[512];
 char CommonLogFile_time[40];
 char cnt[30]="-";
 uint32 sentBytes;
 if(logfile==0) {
     warn_sys("Could not open HTTP log file %s",config_httplogfile);
     return;
 }
 format_address(peer,&p->peer_addr,true);
 format_CommonLogFile_time(CommonLogFile_time,time(0));
 if(PS_getSentBytesCounter(so,&sentBytes)) {
     sprintf(cnt,"%u",sentBytes);
 }
 fprintf(logfile,"%s - %s [%s] \"%s %s HTTP/%s\" %s %s\n",
     peer,
     *p->user!=0?p->user:"-",
     CommonLogFile_time,
     http_req_t_to_string(p->orig_meth),
     p->orig_request_URI,
     p->http_version,
     p->status_code,
     cnt);
 fclose(logfile);
}

static req_parse_status http_header_update(http_header *p, const char *header_line) {
 if(memicmp(header_line,"Host: ",6)==0) {
     strcpy(p->host,header_line+6);
 } else if(memicmp(header_line,"Referer: ",9)==0) {
     strcpy(p->referer,header_line+9);

     if(memicmp(header_line+9,"http://",7)==0) {
          char *slash=strchr(header_line+16,'/');
          if(slash!=NULL) {
               if(memicmp(slash,"/session",8)==0) {
                    sscanf(slash+8,"%X/",&p->referer_sessionID);
               }
          }
     }

     /*a blind fix ...*/
     if(memicmp(header_line+9,"https://",8)==0) {
          char *slash=strchr(header_line+17,'/');
          if(slash!=NULL) {
               if(memicmp(slash,"/session",8)==0) {
                    sscanf(slash+8,"%X/",&p->referer_sessionID);
               }
          }
     }

     if(memicmp(header_line+9,"/session",8)==0) {
          sscanf(header_line+17,"%X/",&p->referer_sessionID);
     }

 } else if(memicmp(header_line,"Authorization: ",15)==0) {
     if(memicmp(header_line+15,"basic ",6)==0) {
          char basic_cookie[256];
          char user[50];
          char pass[80];
          char *colon;
          int len=strlen(header_line+21);

          if(len%4!=0) {
               return req_parse400bad;
          }
          base64decode(header_line+21,basic_cookie,len);
          colon=strchr(basic_cookie,':');
          if(colon==0) {
               p->auth=auth_failed_passwd;
          } else {
               memcpy(user,basic_cookie,colon-basic_cookie);
               user[colon-basic_cookie]=0;
               strcpy(p->user,user);
               strcpy(pass,colon+1);
               strcpy(p->passwd,pass);
               p->auth=http_authenticate(user,pass, inet_ntoa(((struct sockaddr_in*)&p->peer_addr)->sin_addr) ,&p->uid);
          }
     } else {
          p->auth=auth_unsup;
     }
 } else if(memicmp(header_line,"Content-type: ",14)==0) {
     strncpy(p->content_type,header_line+14,sizeof(p->content_type)-1);
 } else if(memicmp(header_line,"User-Agent: ",12)==0) {
     strncpy(p->user_agent,header_line+12,sizeof(p->user_agent)-1);
 } else if(memicmp(header_line,"Cookie: ",8)==0) {
     strncpy(p->cookie,header_line+8,sizeof(p->cookie)-1);
 } else if(memicmp(header_line,"Accept: ",8)==0) {
     strncpy(p->accept,header_line+8,sizeof(p->accept)-1);
 } else if(memicmp(header_line,"Content-length: ",16)==0) {
     sscanf(header_line+16,"%ld",&p->content_length);
 } else if(*header_line!=0) {
     char name[TOKSIZE];
     char *s;
     char *q=strchr(header_line,':');
     if(q==0 || q[1]==0) {
          return req_parse400bad;
     }
     strcpy(name,"HTTP_");
     strncpy(name+5,header_line,q-header_line);
     name[q-header_line+5]=0;
     strupr(name);
     for(s=name+5; *s!=0; s++) {
          if(*s=='-') {
               *s='_';
          }
     }     
     form_data_add(&p->other_headers,name,q+2,false);
 }
 return req_parse200ok;
}

static void change_to_forward_slashes(char *s) {
 while(*s!=0 && *s!='?') {
     if(*s=='\\') *s='/';
     s++;
 }
}
 
void http_blob_handle_to_uri(uint32 handle, char *uri) {
 uint8 *i=(uint8*)&my_IP_address;
 char s_port[10];

 if(http_port==(is_secure?443:80)) {
     s_port[0]=0;
 } else {
     sprintf(s_port,":%d",http_port);
 }
 sprintf(uri,"http%s://%d.%d.%d.%d%s/blob%08lX",
     (is_secure?"s":""),i[0],i[1],i[2],i[3],s_port,handle);
}

static req_parse_status http_parse_request(const char *line, http_header *phd) {
 const char *p1, *p2;
 char _uri[TOKSIZE];
 char *uri=_uri;

 p1=strchr(line,' ');
 if(p1==0) {
     return req_parse400bad;
 }
 p1++;
 p2=strchr(p1,' ');
 if(p2==0) {
     return req_parse400bad;
 }
 while(p1!=p2) {
     *uri++=*p1++;
 }
 *uri=0;
 p2++;
 if(memicmp(p2,"http/",5)!=0) {
     return req_parse400bad;
 }
 phd->full_messages= (p2[5]!='0');
 strncpy(phd->http_version,p2+5,sizeof(phd->http_version));
 phd->http_version[sizeof(phd->http_version)-1]=0;
 if(memicmp(_uri,"/session",8)==0) {
     char *p;
     p=strchr(_uri+8,'/');
     if(p!=0) {
          sscanf(_uri+8,"%X/",&phd->sessionID);
          strcpy(phd->request_URI,p);
          if(!session_valid(phd->sessionID,&phd->peer_addr)) {
               return req_parse404noses;
          }
     } else {
          strcpy(phd->request_URI,_uri);
     }
 } else {
      strcpy(phd->request_URI,_uri);
 }
 change_to_forward_slashes(phd->request_URI);
 strcpy(phd->orig_request_URI,phd->request_URI);

 phd->querystring = strchr(phd->request_URI, '?');
 if(phd->querystring != NULL) phd->querystring++;

 return req_parse200ok;
}

static http_req_t http_which_method(const char *line) {
 if(memicmp(line,"GET",3)==0) {
     return http_get;
 } else if(memicmp(line,"PUT",3)==0) {
     return http_put;
 } else if(memicmp(line,"POST",3)==0) {
     return http_post;
 } else if(memicmp(line,"HEAD",3)==0) {
     return http_head;
 } else if(memicmp(line,"OPTIONS",7)==0) {
     return http_options;
 } else if(memicmp(line,"DELETE",6)==0) {
     return http_delete;
 } else if(memicmp(line,"TRACE",6)==0) {
     return http_trace;
 } else {
     return http_unknown;
 }
}

/* Sessions */

realm_data *realm_data_make(void* data, void (*destroy)(void*)) {
 realm_data *p=(realm_data*)malloc(sizeof(realm_data));
 if(p!=0) {
     p->data=data;
     p->destroy=destroy;
 }
 return p;
}

void session_initialize() {
 sessions=&_sessions;
 sessions->next=sessions->prev=sessions;
 session_num=0;
}

static uint32 session_new_ID(const char *user, struct sockaddr *peer) {
 int i;
 uint32 sessionID=random_next();
 for(i=0; i<MAXSESSIONS; i++,sessionID=random_next()) {
     if(!critical_enter()) {
          return 0;
     }
     if(find_session_cell(sessionID,0,false)==0) {
          make_session_cell(sessionID,user,peer);
          if(!critical_leave()) {
               return 0;
          }
          return sessionID;
     }
 }
 if(!critical_leave()) {
      return 0;
 }
 return 0;     /* Should never get here */
}

static bool session_valid(uint32 sessionID, struct sockaddr *peer) {
 bool ret;
 if(!critical_enter()) {
     return false;
 }
 ret=find_session_cell(sessionID,peer,false)!=0;
 if(!critical_leave()) {
     return false;
 }
 return ret;
}

static void session_grab(uint32 sessionID, int incr) {
 session_cell *p;
 if(!critical_enter()) {
     return;
 }
 p=find_session_cell(sessionID,0,false);
 p->in_use+=incr;
 if(!critical_leave()) {
     return;
 }
}

static void session_reap() {
 session_cell **pp;
 double gap;
 time_t now=time(0);
 if(last_aging==0) {
     last_aging=now;
 }
 if(last_aging==now) {
     return;
 }
 last_aging=now;
 if(!critical_enter()) {
     return;
 }
 pp=&(sessions->next);
 if(*pp!=sessions) {
     for(;*pp!=sessions; pp=&((*pp)->next)) {
          if((*pp)->in_use!=0) {
               (*pp)->accessed=now;
               continue;
          }
          gap=difftime(now,(*pp)->accessed);
          if(gap>60.0*config_httpsessiontimeout) {
               session_cell *to_free=*pp;
               (*pp)->next->prev=(*pp)->prev;
               *pp=(*pp)->next;
               session_data_reinit(&to_free->data);
               checked_free(to_free);
          }
     }
 }
 if(!critical_leave()) {
     return;
 }
}

void session_close(uint32 sessionID) {
 session_cell **pp;
 if(!critical_enter()) {
     return;
 }
 pp=&(sessions->next);
 if(*pp!=sessions) {
     for(;*pp!=sessions; pp=&((*pp)->next)) {
          if((*pp)->sessionID==sessionID) {
               session_cell *to_free=*pp;
               (*pp)->next->prev=(*pp)->prev;
               *pp=(*pp)->next;
               session_data_reinit(&to_free->data);
               checked_free(to_free);
          }
     }
 }
 if(!critical_leave()) {
     return;
 }
}

static bool session_set_user(uint32 sessionID, const char *user) {
 session_cell *p;
 if(!critical_enter()) {
     return false;
 }
 p=find_session_cell(sessionID,0,false);
 if(p!=0) {
     strcpy(p->user,user);
 }
 if(!critical_leave()) {
     return false;
 }
 return p!=0;
}


realm_data *session_get_realm_data(uint32 sessionID, const char *realm) {
 session_cell *p;
 realm_data *ret;
 if(!critical_enter()) {
     return 0;
 }
 p=find_session_cell(sessionID,0,true);
 if(p==0) {
     return 0;
 }
 ret=session_data_get_realm_data(&p->data,realm);
 if(!critical_leave()) {
     return 0;
 }
 return ret;
}

bool session_set_realm_data(uint32 sessionID, const char *realm, realm_data *data) {
 session_cell *p;
 if(!critical_enter()) {
     return false;
 }
 p=find_session_cell(sessionID,0,false);
 if(p==0) {
     return false;
 }
 session_data_set_realm_data(&p->data,realm,data);
 if(!critical_leave()) {
     return false;
 }
 return true;
}

void session_html_dump(PS_SOCKET so, const char *delete_session_URL_prefix) {
 session_cell *p;
 if(!critical_enter()) {
     return;
 }
 p=sessions->next;
 if(p!=sessions) {
     PS_printf(so,"<table border bgcolor=\"CCFFFF\"><tr bgcolor=\"33FFFF\">"
          "<td align=\"right\"><b>Id</b></td>"
          "<td><b>From</b></td>"
          "<td><b>On since</b></td>"
          "<td><b>Idle since</b></td>"
          "<td><b>Initial User</b></td>"
          "<td><b>User</b></td>"
          "<td><b>Realms</b></td>");
     if(delete_session_URL_prefix!=0) {
          PS_printf(so,"<td><b>Remove</b></td>");
     }          
     PS_printf(so,"</tr>\n");
     for(;p!=sessions; p=p->next) {
          char onsince[100];
          char accessed[100];
          char peer[100];
          format_time(onsince,p->created);
          if(p->in_use) {
               strcpy(accessed,"<font color=\"#FF0000\">");
               format_time(accessed+22,time(0));
               strcat(accessed,"</font>");
          } else {
               format_time(accessed,p->accessed);
          }
          format_address(peer,&p->peer_addr,true);
          PS_printf(so,"<tr><td align=\"right\">%X</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td>",
               p->sessionID,peer,onsince,accessed,p->init_user,p->user);
          if(p->data.realms==0) {
               PS_printf(so,"<td>&nbsp;</td><tr>\n");
          } else {
               realm_link *r=p->data.realms;
               PS_printf(so,"<td>");
               for(;r!=0; r=r->next) {
                    PS_printf(so,"%s ",r->realm);
               }
               if(delete_session_URL_prefix!=0) {
                    PS_printf(so,"<td><a href=\"%s%X\">Remove</a></td>",delete_session_URL_prefix,p->sessionID);
               }
               PS_printf(so,"</td><tr>\n");
          }
     }
     PS_printf(so,"</table>\n");
 } else {
     PS_printf(so,"There are no sessions<br>\n");
 }

 if(!critical_leave()) {
     return;
 }
}

static session_cell *make_session_cell(uint32 sessionID, const char *user, struct sockaddr *peer) {
 session_cell *p;
 session_cell *last=sessions->prev;
 if(session_num==MAXSESSIONS) {
     p=sessions->next;
     /* Detach from front */
     sessions->next=p->next;
     sessions->next->prev=sessions;
     /* Re-Initialize data */
     session_data_reinit(&p->data);
     p->sessionID=sessionID;
 } else {
     p=(session_cell*)checked_malloc(sizeof(session_cell));
     if(p==0) {
          return 0;
     }
     /* Initialize data */
     session_data_init(&p->data);
     p->sessionID=sessionID;
 }
 p->in_use=0;
 p->accessed=p->created=time(0);
 strcpy(p->init_user,user);
 strcpy(p->user,user);
 memcpy(&p->peer_addr,peer,sizeof(struct sockaddr));
 /* Add to back */
 sessions->prev=p;
 p->next=sessions;
 p->prev=last;
 last->next=p;
 return p;
}

static session_cell *find_session_cell(uint32 sessionID, struct sockaddr *peer, bool boost) {
 session_cell *p=sessions->next;
 for(p=sessions->next;p!=sessions;p=p->next) {
     if(sessionID==p->sessionID && (peer==0 || same_host(&p->peer_addr,peer))) {
          if(boost) {
               session_cell *last;
               p->accessed=time(0);
               /* Detach from current position */
               p->prev->next=p->next;
               p->next->prev=p->prev;
               /* Add to back */
               last=sessions->prev;
               sessions->prev=p;
               p->next=sessions;
               p->prev=last;
               last->next=p;
          }
          return p;
     }
 }
 return 0;
}

static void session_data_init(session_data *s) {
 s->realms=0;
}

static void session_data_reinit(session_data *s) {
 realm_link *p=s->realms;
 while(p!=0) {
     realm_link *n=p->next;
     checked_free(p->realm);
     p->object->destroy(p->object->data);
     checked_free(p->object);
     p=n;
 }
 session_data_init(s);
}

static realm_link *session_data_find_realm_link(session_data *s, const char *realm) {
 realm_link *p=s->realms;
 while(p!=0) {
     if(strcmp(p->realm,realm)==0) {
          return p;
     }
     p=p->next;
 }
 p=(realm_link*)checked_malloc(sizeof(realm_link));
 p->realm=checked_strdup(realm);
 p->object=0;
 p->next=s->realms;
 s->realms=p;
 return p;
}

static realm_data *session_data_get_realm_data(session_data *s, const char *realm) {
 realm_link *p=session_data_find_realm_link(s,realm);
 if(p!=0) {
     return p->object;
 } else {
     return 0;
 }
}

static void session_data_set_realm_data(session_data *s, const char *realm, realm_data *data) {
 realm_link *p=session_data_find_realm_link(s,realm);
 if(p!=0) {
     p->object=data;
 } 
}

static bool filetype_is_executable(const char *ext) {
 const char *p=config_execext;
 int len=strlen(ext);
 while(*p!=0) {
     while(*p!=0 && isascii(*p) && isspace(*p)) p++;
     if(*p==0) {
          break;
     }
     if(memicmp(ext,p,len)==0) {
          return true;
     }
     while(*p!=0 && isascii(*p) && !isspace(*p)) p++;
 }
 return false;
}

static bool mimetype_run_command_string(const char *mimetype, const char *fname,
     char *cmdbuffer) {
 char m[513], *q;
 const char *p=config_binpath;
 strcpy(m,mimetype);
 strlwr(m);
 for(q=m; *q!=0; q++) {
     if(*q=='/') {
          *q='_';
     }
 }     
 
 sprintf(cmdbuffer,"%s%s-handler",p,m);
 if(access(cmdbuffer,0)!=0) {
     return false;
 }
 sprintf(cmdbuffer,"%s%s-handler %s",p,m,fname);
 return true;
}

#ifdef NT
/* Registry Bull**it again! One has to do it twice to produce the value
   and it still says it failed! */
static void MyRegQueryValue(
  HKEY hKey,       // handle to key to query
  LPCTSTR lpSubKey,                   // name of subkey to query
  LPTSTR lpValue,  // buffer for returned string
  PLONG lpcbValue  // receives size of returned string
  ) {
  if(RegQueryValue(hKey,lpSubKey,lpValue,lpcbValue)!=ERROR_SUCCESS) {
     RegQueryValue(hKey,lpSubKey,lpValue,lpcbValue);
  }
}

static bool filetype_open_command_string(const char *ext, const char *fname,
     char *cmdbuffer) {
 int sz;
 char dbkey[513];
 char buffer[513];
 HKEY start=HKEY_LOCAL_MACHINE;
 char *p;

 /* Query filetype */
 buffer[0]=0;
 sprintf(dbkey,"SOFTWARE\\CLASSES\\.%s",ext);
 MyRegQueryValue(start, dbkey, buffer, &sz);
 if(buffer[0]==0) {
     return false;
 }

 /* Query open command string */
 sprintf(dbkey,"SOFTWARE\\CLASSES\\%s\\Shell\\Open\\Command",buffer);
 buffer[0]=0;
 MyRegQueryValue(start, dbkey, buffer, &sz);
 if(buffer[0]==0) {
     return false;
 }

 /* use the filename in the place of %0 or %1... */
 p=buffer;
 while(*p!=0) {
     if(*p=='%' && (p[1]=='0' || p[1]=='1')) {
          strcat(cmdbuffer,fname);
          cmdbuffer+=strlen(fname);
          p+=2;
          continue;
     }
     *cmdbuffer++=*p++;
 }

 return true;
}
#endif

static int indirect_strcmp(const void *p1, const void *p2) {
 return strcmp(*(const char**)p1,*(const char**)p2);
}

/* All additional headers are put in the environment preceded by "HTTP_"
   and are strdup'ed to be distinguishable */
static void free_strduped_env(char **env) {
 for(;*env!=0;env++) {
     if(memcmp(*env,"HTTP_",5)==0) {
          checked_free(*env);
     }          
 }
}

static bool is_nph_cgi(const char *cmdline) {
 const char *last=cmdline;
 for(; *cmdline!=0 && *cmdline!=' '; cmdline++) {
     if(*cmdline=='/') {
          last=cmdline;
     }
 }
 return memcmp(last,"nph-",4)==0;
}

typedef struct cgi_state {
 char session_data[TOKSIZE];
} cgi_state;

static cgi_state *cgi_state_construct() {
 cgi_state *p=(cgi_state*)malloc(sizeof(cgi_state));
 if(p!=0) {
     p->session_data[0]=0;
 }
 return p;
}

static void cgi_state_destruct(void *pv) {
}

typedef struct {
 PS_SOCKET so;
 const char *realm;
 char *buffer;
 http_header* phd;
 char *cmdline;
} produce_args;

static hdl_t http_async_produce_external_content_aux2(produce_args *ha);

static void http_async_produce_external_content_aux1(void *pv) {
 hdl_t ret;
 produce_args *ha=(produce_args*)pv;
 ret=http_async_produce_external_content_aux2(ha);
 if(!http_status_code(ha->so,ha->phd,ret)) {
     http_404_not_found(ha->so,ha->phd);
 }
 http_log_request(ha->phd,ha->so);
 PS_closesocket(ha->so);
 PS_release(ha->so);
 http_header_deinit(ha->phd);
 checked_free(ha->phd);
 checked_free(ha->cmdline);
 checked_free(ha);
}

hdl_t http_async_produce_external_content(
     PS_SOCKET so,
     char *buffer,
     http_header *phd,
     const char *cmdline) {
 sint32 id;
 produce_args *ha=checked_malloc(sizeof(produce_args));
 http_header *phd2=checked_malloc(sizeof(http_header));
 if(ha==0 || phd2==0) {
     checked_free(ha);
     checked_free(phd2);
     return hdl_error;
 }
 http_header_copy(phd2,phd);
 ha->so=so;
 ha->realm="/upload";
 ha->buffer=buffer;
 ha->phd=phd2;
 ha->cmdline=checked_strdup(cmdline);
 spawn_cgi(http_async_produce_external_content_aux1,ha,&id);
 if(children_are_threads()) {
     return hdl_defer;
 } else {
     form_data_abandon_files(&phd2->form);
     http_header_deinit(phd2);	
     checked_free(ha);
     checked_free(phd2);
     return hdl_defer_deinit;
 }
}

static hdl_t http_async_produce_external_content_aux2(produce_args *ha) {
 PS_SOCKET so=ha->so;
 const char *realm=ha->realm;
 /* Do not use ha->buffer. It is shared. */
 char buffer[HTTP_BUFFER_SZ];
 http_header* phd=ha->phd;
 char *cmdline=ha->cmdline;
 return http_produce_external_content(so,buffer,phd,cmdline);
}

#define ENVNUM 50

hdl_t http_produce_external_content(
     PS_SOCKET so,
     char *buffer,
     http_header *phd,
     const char *cmdline) {
 char cmd_line[TOKSIZE];
 PIPE medium[2];
 char *env[ENVNUM];
 char *argv[20];
 char env_query_string[TOKSIZE+100];
 char env_server_software[100];
 char env_server_name[513];
 char env_gateway_interface[]="GATEWAY_INTERFACE=CGI/1.1";
 char env_server_protocol[]="SERVER_PROTOCOL=HTTP/1.0";
 char env_server_port[20];
 char env_script_name[TOKSIZE];
 char env_auth_type[30];
 char env_request_method[100];
 char env_remote_addr[70];
 char env_path_info[530];
 char env_path_translated[530];
 char env_path[1024];
 char env_http_referer[TOKSIZE+100];
 char env_remote_user[70];
 char env_remote_password[70];
 char env_content_type[70];
 char env_http_user_agent[256];
 char env_http_cookie[TOKSIZE+20];
 char env_http_accept[256];
 char env_content_length[30];
 char env_session_id[30];
 char env_x_session_data[TOKSIZE];
#if defined(NT)
 char env_systemroot[512];
 char env_systemdrive[64];
#endif
 int ienv=0, ienv2;
 int iarg=0;
 int iadd;
 char *p;
 sint32 length;
 uint32 env_length=0;
 form_element *e;
 form_data cgi_response_headers;
 bool nph=is_nph_cgi(cmdline);
 realm_data *rdata=0;
 cgi_state *state;

 strcpy(cmd_line,cmdline);

 env_query_string[0]=0;
 if(phd->meth==http_get && !form_data_empty(&phd->form)) {
     char query_string[TOKSIZE];
     form_data_url_encode(&phd->form,query_string);
     sprintf(env_query_string,"QUERY_STRING=%s",query_string);
 }
 if(phd->meth==http_get && phd->form.isindex_query[0]!=0) {
     char query_string[TOKSIZE];
     form_data_url_encode(&phd->form,phd->form.isindex_query);
     sprintf(env_query_string,"QUERY_STRING=%s",phd->form.isindex_query);
     strcat(cmd_line," ");
     strcat(cmd_line,query_string);
 }

 sprintf(env_server_protocol,"SERVER_PROTOCOL=%s",
     (is_secure?"HTTPS/1.0":"HTTP/1.0"));
 sprintf(env_server_software,"SERVER_SOFTWARE=%s",server_token);
 sprintf(env_server_name,"SERVER_NAME=%s",my_hostname);
 sprintf(env_server_port,"SERVER_PORT=%d",http_port);
 sprintf(env_request_method,"REQUEST_METHOD=%s",http_req_t_to_string(phd->meth));
 sprintf(env_path_info,"PATH_INFO=%s",phd->path_info);
 sprintf(env_path_translated,"PATH_TRANSLATED=%s%s",config_httprootdir,
     phd->path_info);
 sprintf(env_path,"PATH=%s",getenv("PATH"));
 sprintf(env_http_referer,"HTTP_REFERER=%s",phd->referer);
 sprintf(env_remote_addr,"REMOTE_ADDR=");
 format_address(env_remote_addr+12,&phd->peer_addr,true);
 sprintf(env_content_type,"CONTENT_TYPE=%s",phd->content_type);
 sprintf(env_content_length,"CONTENT_LENGTH=%d",phd->content_length);
 sprintf(env_http_user_agent,"HTTP_USER_AGENT=%s",phd->user_agent);
 if(phd->cookie[0]!=0) {
     sprintf(env_http_cookie,"HTTP_COOKIE=%s",phd->cookie);
 }
 sprintf(env_http_accept,"HTTP_ACCEPT=%s",phd->accept);
 for(iadd=0; iadd<additional_env_vars_nb; iadd++) {
     env[ienv++]=additional_env_vars[iadd];
 }
 sprintf(env_script_name,"SCRIPT_NAME=%s",phd->request_URI);

 rdata=session_get_realm_data(phd->sessionID,phd->request_URI);
 if(rdata!=0) {
     /* Guard against indirect calls from persistent handlers */
     if(rdata->destroy==cgi_state_destruct) {
          state=(cgi_state*)rdata->data;
          sprintf(env_x_session_data,"X_SESSION_DATA=%s",state->session_data);
     } else {
          rdata=0;
     }          
 }

 if(phd->auth==auth_ok) {
     sprintf(env_auth_type,"AUTH_TYPE=Basic");
     env[ienv++]=env_auth_type;
 }
 env[ienv++]=env_content_length;
 env[ienv++]=env_content_type;
 env[ienv++]=env_gateway_interface;
 env[ienv++]=checked_strdup(env_http_accept);
 if(phd->cookie[0]!=0) {
     env[ienv++]=env_http_cookie;
 }
 env[ienv++]=checked_strdup(env_http_referer);
 env[ienv++]=checked_strdup(env_http_user_agent);
 env[ienv++]=env_path;
 env[ienv++]=env_path_info;
 env[ienv++]=env_path_translated;
 if(env_query_string[0]!=0) {
     env[ienv++]=env_query_string;
 }
 if(phd->auth==auth_ok) {
     sprintf(env_remote_password,"REMOTE_PASSWORD=%s",phd->passwd);
     env[ienv++]=env_remote_password;
     sprintf(env_remote_user,"REMOTE_USER=%s",phd->user);
     env[ienv++]=env_remote_user;
 }
 env[ienv++]=env_request_method;
 env[ienv++]=env_script_name;
 env[ienv++]=env_server_name;
 env[ienv++]=env_server_port;
 env[ienv++]=env_server_protocol;
 env[ienv++]=env_server_software;
 if(phd->sessionID!=0) {
     sprintf(env_session_id,"SESSION_ID=%d",phd->sessionID);
     env[ienv++]=env_session_id;
 }
#if defined(NT)
 sprintf(env_systemroot,"SYSTEMROOT=%s",getenv("systemroot"));
 sprintf(env_systemdrive,"SYSTEMDRIVE=%s",getenv("systemdrive"));
 env[ienv++]=env_systemroot;
 env[ienv++]=env_systemdrive;
#endif
 if(rdata!=0) {
     env[ienv++]=env_x_session_data;
 }     

 /* Count bytes used for environment */
 for(ienv2=0; ienv2<ienv; ienv2++) {
     env_length+=strlen(env[ienv2])+1;  
 }
 env_length++;
 if(env_length>ENVSIZE) {
     /* ENVSIZE exceeded */
     return hdl_error;
 }

 /* Add other headers from a marked position onwards */
 /* Take care to respect ENVSIZE and ENVNUM */
 ienv2=ienv;
 e=phd->other_headers.elements;
 while(e!=0) {
     char n[TOKSIZE];
     if(env_length+strlen(e->name)+strlen(e->data)+1>ENVSIZE) {
          break;
     }
     sprintf(n,"%s=%s",e->name,e->data);
     env[ienv++]=checked_strdup(n);
     env_length+=strlen(e->name)+strlen(e->data)+1;
     e=e->next;
     if(ienv==ENVNUM-1) {
          break;
     }
 }
 env[ienv++]=0;
 
 /* Sort environment */
 qsort(env,ienv-1,sizeof(char*),indirect_strcmp);
 
 /* Split arguments */
 p=strtok(cmd_line," \n");
 if(p!=0) {
     do {
          if(iarg==sizeof(argv)/sizeof(char*)) {
               free_strduped_env(env);
               return hdl_error;
          }
          argv[iarg++]=p;
     } while(p=strtok(0," \n"));
 }
 argv[iarg]=0;

 if(!pipexecve(medium,argv[0],env,argv)) {
     free_strduped_env(env);
     return hdl_error;
 }

 /* Feed input to pipe directly */
 length=phd->content_length;
 while(length>0) {
     int blocklen=length;
     if(blocklen>CGI_BUFFER_SZ) {
          blocklen=CGI_BUFFER_SZ;
     }
     if(PS_recv_all(so,buffer,blocklen)<0) {
          pipe_close(medium[0]);
          pipe_close(medium[1]);
          free_strduped_env(env);
          return hdl_error;
     }
     if(pipe_write(medium[1],buffer,blocklen)!=blocklen) {
          pipe_close(medium[0]);
          pipe_close(medium[1]);
          free_strduped_env(env);
          return hdl_error;
     }
     length-=blocklen;
 }
 pipe_close(medium[1]);

 if(!nph) {
     /* Unless NPH, read and parse response headers */
     form_element *e;
     form_data_init(&cgi_response_headers);
     
     do {
          char *q, *r;
          char mime_type[70];
          if(!pipe_read_line(medium[0], buffer, CGI_BUFFER_SZ)) {
               pipe_close(medium[0]);
               form_data_deinit(&cgi_response_headers);
               free_strduped_env(env);
               return hdl_error;
          }
          if(*buffer==0) {
               break;
          }
          
          q=strchr(buffer,':');
          if(q==0 || q[1]==0) {
               pipe_close(medium[0]);
               form_data_deinit(&cgi_response_headers);
               free_strduped_env(env);
               return hdl_error;
          }
          *q=0;
          q+=2;
          if(stricmp(buffer,"Content-type")==0) {
               r=strchr(q,'/');
               if(r==0) {
                    /* Extension specified. Deduce MIME type */
                    http_map_ext_to_mime_type(q, mime_type);
                    q=mime_type;
               }
          }
          form_data_add(&cgi_response_headers,buffer,q,false);
          
     } while(1);

     if(form_data_element(&cgi_response_headers, "X-Session-Data", buffer, 0)) {
          if(phd->sessionID==0) {
               pipe_close(medium[0]);
               form_data_deinit(&cgi_response_headers);
               free_strduped_env(env);
               return hdl_newses;
          }
          if(rdata==0) {
               state=cgi_state_construct();
               if(state!=0) {
                    rdata=realm_data_make(state,cgi_state_destruct);
               }
               if(rdata==0) {
                    pipe_close(medium[0]);
                    form_data_deinit(&cgi_response_headers);
                    free_strduped_env(env);
                    return hdl_error;
               }
               session_set_realm_data(phd->sessionID,phd->request_URI,rdata);
          }
          state=(cgi_state*)rdata->data;
          strcpy(state->session_data,buffer);
     }
     
     /* If redirection is required, redirect to internal handler, or URL */
     if(form_data_element(&cgi_response_headers, "Location", buffer, 0)) {
          if(*buffer=='/' && strchr(buffer,'#')==0) {           
               strcpy(phd->request_URI,buffer);
               pipe_close(medium[0]);
               form_data_deinit(&cgi_response_headers);
               free_strduped_env(env);
               return hdl_redirect;
          } else {               
               PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n");
               strcpy(phd->status_code,"302");
               if(*buffer=='/') { /* virtual path with '#' */
                    char s_base[512];
                    format_base_URI(s_base,phd);
                    if(phd->sessionID!=0) {
                         PS_printf(so,"Location: %s/session%X%s\r\n\r\n",
                              s_base,phd->sessionID,buffer);
                    } else {
                         PS_printf(so,"Location: %s/%s\r\n\r\n",
                              s_base,buffer);
                    }
               } else {
                    PS_printf(so,"Location: %s\r\n\r\n",buffer);
               }     
               return hdl_ok;
          }                    
     }
     if(!form_data_element(&cgi_response_headers, "Status", buffer, 0)) {
          strcpy(buffer,"200 OK");
     }
     memcpy(phd->status_code,buffer,3);
     phd->status_code[3]=0;
     PS_printf(so,"HTTP/1.0 %s\r\n",buffer);

     /* Send headers (Script-Control is ignored)*/
     e=cgi_response_headers.elements;
     while(e!=0) {
          if(stricmp(e->name,"Status")!=0
          && strcmp(e->name,"Script-Control")!=0) {
               PS_printf(so,"%s: %s\r\n",e->name,e->data);
          }
          e=e->next;
     }
     PS_printf(so,"\r\n");

 }      

 PS_resetSentBytesCounter(so);
 /* Feed output to socket */
 do {
     length=pipe_read(medium[0],buffer,CGI_BUFFER_SZ);
     if(length<0) {
          pipe_close(medium[0]);
          form_data_deinit(&cgi_response_headers);
          free_strduped_env(env);
#if defined(NT)
          /* NT: no 0 for end-of-input :-< */
          return hdl_ok;
#else
          return hdl_error;
#endif
     }
     if(length==0) {
          break;
     }
     if(PS_send_all(so,buffer,length)<0) {
          pipe_close(medium[0]);
          form_data_deinit(&cgi_response_headers);
          free_strduped_env(env);
          return hdl_error;
     }
 } while(length!=0);

 pipe_close(medium[0]);
 form_data_deinit(&cgi_response_headers);
 free_strduped_env(env);
 
 return hdl_ok;

}

const char *http_req_t_to_string(http_req_t meth) {
 switch(meth) {
     case http_unknown:  return "UNKNOWN";
     case http_get:      return "GET";
     case http_head:     return "HEAD";
     case http_options:  return "OPTIONS";
     case http_post:     return "POST";
     case http_put:      return "PUT";
     case http_delete:   return "DELETE";
     case http_trace:    return "TRACE";
 }
 return "?"; /* not reached */
}

void http_send_file_data(PS_SOCKET so, const char *fname, char *buffer) {
    PS_send_file(so,fname,buffer,HTTP_BUFFER_SZ);
}

void http_send_file_data_range(PS_SOCKET so, const char *fname, char *buffer,
     uint32 offset_s, uint32 offset_e) {
    PS_send_file_range(so,fname,buffer,HTTP_BUFFER_SZ, offset_s, offset_e);
}

static void http_map_ext_to_mime_type(const char *ext, char *mime_type) {
 FILE *f;
 if(*config_mimetypesfile!=0 && (f=fopen(config_mimetypesfile,"r"))!=0) {
     char line[256];
     char *p;

     while (fgets(line,sizeof(line),f)!=0) {
          p=strtok(line," \t\n");    /* This delimits mime type with '\0' */

          while((p=strtok(0," \t\n,"))!=0) {
#if defined(NT)
               if(stricmp(ext,p)==0) {
#else
               if(strcmp(ext,p)==0) {
#endif
                    strcpy(mime_type,line);
                    fclose(f);
                    return;
               }
          }
     }
     fclose(f);
 }
 strcpy(mime_type,"application/octet-stream");
}

/* Called after normalizing path */
static bool directory_traversal_path(const char *s) {
 const char *p=strrchr(s,'/');
 if(p!=0) {
     return (strcmp(p+1,".")==0) || (strcmp(p+1,"..")==0);
 }
 return false;
}

static int safe_stat(const char *fname, struct stat *pst) {
 char fname2[1024];
 int len=strlen(fname);
 if(fname[len-1]=='/') {
     strcpy(fname2,fname);
     fname2[len-1]=0;
     fname=fname2;
 }
 if(directory_traversal_path(fname)) {
     errno=EACCES;
     return -1;
 }
 return stat(fname,pst);
}

#if defined(NO_THREADS)
static hdl_t file_handler_aux2(handler_args *ha);

static void file_handler_aux1(void *pv) {
 hdl_t ret;
 handler_args *ha=(handler_args*)pv;
 ret=file_handler_aux2(ha);
 if(!http_status_code(ha->so,ha->phd,ret)) {
     http_404_not_found(ha->so,ha->phd);
 }
 http_log_request(ha->phd,ha->so); /*mc: check that this is also done on UNIX*/
 PS_closesocket(ha->so);
 PS_release(ha->so);
 http_header_deinit(ha->phd);
 checked_free(ha->phd);
 checked_free(ha);
}

hdl_t file_handler_nothreads(PS_SOCKET so, const char *realm, char *buffer, http_header* phd) {
 hdl_t ret;
 handler_args *ha=checked_malloc(sizeof(handler_args));
  if(ha==NULL) {
     checked_free(ha);
     return hdl_error;
 }
 ha->so=so;
 ha->realm=realm;
 ha->buffer=buffer;
 ha->phd=phd;
 ret = file_handler_aux2(ha);
 checked_free(ha);
 return ret;
}

hdl_t file_handler(PS_SOCKET so, const char *realm, char *buffer, http_header* phd) {
 sint32 id;
 handler_args *ha=checked_malloc(sizeof(handler_args));
 http_header *phd2=checked_malloc(sizeof(http_header));
 if(ha==0 || phd2==0) {
     checked_free(ha);
     checked_free(phd2);
     return hdl_error;
 }
 http_header_copy(phd2,phd);
 ha->so=so;
 ha->realm=realm;
 ha->buffer=buffer;
 ha->phd=phd2;
 spawn_cgi(file_handler_aux1,ha,&id);
 if(children_are_threads()) {
     return hdl_defer;
 } else {
     http_header_deinit(phd2);	
     checked_free(ha);
     checked_free(phd2);
     return hdl_defer_deinit;
 }
}

static hdl_t file_handler_aux2(handler_args *ha) {
 PS_SOCKET so=ha->so;
 const char *realm=ha->realm;
 /* Do not use ha->buffer. It is shared. */
 char buffer[HTTP_BUFFER_SZ];
 http_header* phd=ha->phd;
#else
hdl_t file_handler_nothreads(PS_SOCKET so, const char *realm, char *buffer, http_header* phd) {
    return file_handler(so, realm, buffer, phd);
}
hdl_t file_handler(PS_SOCKET so, const char *realm, char *buffer, http_header* phd) {
#endif
 char mime_type[81];
 char fname[TOKSIZE];
 char *p;
 const char *ext;
 char mtime[80];
 struct stat st;
 enum { dir_root, dir_shared };
 int dir=(config_httprootdir[0]==0?dir_shared:dir_root);
 char range[TOKSIZE];
 uint32 range_start, range_end, total_size;

 if((phd->meth==http_get || phd->meth==http_head)
 && strchr(phd->request_URI,'?')!=0) {
     if(!handler_read_form_data(phd, so)) {
          /* Error Status code returned already */
          return hdl_ok;
     }
 }

 /* Find component in URI that is a regular file */
 /* The rest is PATH_INFO */
 p=phd->request_URI;
 while((p=strchr(p+1,'/'))!=0) {
     if(dir==dir_root) {
          strcpy(fname,config_httprootdir);
          strncpy(fname+strlen(config_httprootdir),phd->request_URI+1,p-phd->request_URI);
          fname[p-phd->request_URI+strlen(config_httprootdir)]=0;
          if(safe_stat(fname,&st)<0) {
               if(errno==ENOENT) {
                    dir=dir_shared;
               } else {
                    return hdl_error;
               }
          } else {
               if(st.st_mode & S_IFDIR) {
                    continue;
               } else if(st.st_mode & S_IFREG) {
                    break;
               } else {
                    return hdl_error;
               }
          }
     }
     if(dir==dir_shared) {
          strcpy(fname,config_httpshareddir);
          strncpy(fname+strlen(config_httpshareddir),phd->request_URI+1,p-phd->request_URI);
          fname[p-phd->request_URI+strlen(config_httpshareddir)]=0;
          if(safe_stat(fname,&st)<0) {
               if(errno==ENOENT) {
                    return hdl_sep;
               } else {
                    return hdl_error;
               }
          } else {
               if(st.st_mode & S_IFDIR) {
                    continue;
               } else if(st.st_mode & S_IFREG) {
                    break;
               } else {
                    return hdl_error;
               }
          }
     }
 }

 if(p!=0) {
     strcpy(phd->path_info,p);
     *p=0;
 }
 if(dir==dir_root) {
     strcpy(fname,config_httprootdir);
 } else {
     strcpy(fname,config_httpshareddir);
 }
 strcat(fname,phd->request_URI+1);
 if(p==0) {
     if(safe_stat(fname,&st)<0) {
          if(errno==ENOENT) {
               if(dir==dir_root) {
                    dir=dir_shared;
                    strcpy(fname,config_httpshareddir);
                    strcat(fname,phd->request_URI+1);
                    if(safe_stat(fname,&st)<0) {
                         if(errno==ENOENT) {
                              return hdl_sep;
                         } else {
                              return hdl_error;
                         }
                    }
               } else {
                    if(errno==ENOENT) {
                         return hdl_sep;
                    } else {
                         return hdl_error;
                    }
               }
          } else {
               return hdl_error;
          }
     }
 }

 if(st.st_mode & S_IFDIR) {
     const char *index_file;
     uint8 *i=(uint8*)&my_IP_address;
     char s_name[255];
     char s_port[10];

     if(phd->meth!=http_get && phd->meth!=http_head) {
          return hdl_meth;
     }

     if(http_port==(is_secure?443:80)) {
          s_port[0]=0;
     } else {
          sprintf(s_port,":%d",http_port);
     }
     if(phd->host!=0) {
          if(strchr(phd->host,':')==0 && s_port[0]==':') {
               /* Ericsson browser does not put :port in Host... */
               sprintf(s_name,"%s%s",phd->host,s_port);
          } else {               
               strcpy(s_name,phd->host);
          }
     } else {
          sprintf(s_name,"%d.%d.%d.%d%s",i[0],i[1],i[2],i[3],s_port);
     }

     if(fname[strlen(fname)-1]!='/') {
          strcat(fname,"/");
     }

     if(phd->request_URI[strlen(phd->request_URI)-1]!='/') {
          strcat(phd->request_URI,"/");
     }

     if(http_demands_wap(phd)) {
          index_file="index.wml";
     } else {
          index_file="index.html";
     }
     
     strcat(fname,index_file);
     if(access(fname,0)!=0) {                     /* If index file not found */
          if(*config_httpshareddir==0) {          /* and HTTPSHAREDDIR */
               return hdl_sep;                    /* exists, */
          }
          strcpy(fname,config_httpshareddir);     /* try there */
          strcat(fname,phd->request_URI+1);
          strcat(fname,index_file);
          if(access(fname,0)!=0) {                /* If not found */
               return hdl_sep;                    /* give up */
          }
     }

     strcat(phd->request_URI,index_file);
     http_redirect(so,phd,"http%s://%s%s\r\n\r\n",
         is_secure?"s":"",s_name,phd->request_URI);
/*
     PS_printf(so,"HTTP/1.0 302 Moved temporarily\r\n");
     strcpy(phd->status_code,"302");
     PS_printf(so,"Location: http%s://%s%s\r\n\r\n",
         is_secure?"s":"",s_name,phd->request_URI);
*/
     return hdl_ok;
 } else if(st.st_mode & S_IFREG) {
     int i;
     ext=strrchr(fname,'.');
     if(ext!=0) {
          ext++;
          http_map_ext_to_mime_type(ext,mime_type);
     } else {
          strcpy(mime_type,"text/plain");
     }

     for(i=0; i<mime_type_handlers_nb; i++) {
          if(stricmp(mime_type,mime_type_handlers[i].mime_type)!=0) {
               continue;
          }
          return mime_type_handlers[i].func(so,buffer,phd,fname);
     }          

     if(st.st_mode & S_IEXEC) {
          return http_produce_external_content(
               so,
               buffer,
               phd,
               fname);
     }

     if(filetype_is_executable(ext)) {
          char cmd[1024];
#ifdef NT
          if(filetype_open_command_string(ext, fname,cmd)) {
               return http_produce_external_content(
                    so,
                    buffer,
                    phd,
                    cmd);
          }
#endif
          if(mimetype_run_command_string(mime_type, fname,cmd)) {
               return http_produce_external_content(
                    so,
                    buffer,
                    phd,
                    cmd);
          }
          return hdl_error;
     }
     
     if(phd->meth!=http_get && phd->meth!=http_head) {
          return hdl_meth;
     }

     format_rfc1123_time(mtime,st.st_mtime);

     if(!form_data_element(&phd->other_headers,"HTTP_RANGE",range,0)) {
          *range=0;
          total_size=st.st_size;
          range_start=0;
          range_end=total_size-1;
     } else {
          int nums[2];
          int lens[2];
          int cnt;
          nums[1]=st.st_size-1;
          if(*range=='-') {
               if(!simparse(range,"s98",nums,lens,&cnt)) {
                    goto parse_error;
               }
               if(-nums[0]>st.st_size) {
                    range_start=0;
                    range_end=st.st_size-1;
               } else {
                    range_start=st.st_size+nums[0];
                    range_end=st.st_size-1;
               }
          } else {
               if(!simparse(range,"98-?98",nums,lens,&cnt)) {
                    goto parse_error;
               }
               if(nums[0]>nums[1]) {
                    goto parse_error;
               }
               if(nums[1]>st.st_size-1) {
                    nums[1]=st.st_size-1;
               }
               range_start=nums[0];
               range_end=nums[1];
          }
     }
     total_size=range_end-range_start+1;

     if(phd->full_messages) {
          const char *resp;
          if(total_size==(uint32)st.st_size) {
               resp="200 OK";
          } else {
               resp="206 Partial Content";
          }
          PS_printf(so,"HTTP/1.0 %s\r\nContent-type: %s\r\n"
               "Pragma: no-cache\r\nConnection: close\r\n"
               "Last-Modified: %s\r\nContent-length: %u\r\n"
               "Server: %s\r\n\r\n",
               resp,mime_type,mtime,total_size,server_token);
          memcpy(phd->status_code,resp,3);
          phd->status_code[3]=0;
     }

     if(phd->meth==http_head) {
          return hdl_ok;
     }

     PS_resetSentBytesCounter(so);
     if(phd->meth==http_head) {
          return hdl_ok;
     }

     http_send_file_data_range(so,fname,buffer,range_start,range_end);
 }

 return hdl_ok;

parse_error:
 PS_printf(so,"HTTP/1.0 406 Not Acceptable\r\nContent-type: text/html\r\nPragma: no-cache\r\n"
     "Connection: close\r\nServer: %s\r\n\r\n",
     server_token);
 PS_printf(so,"<html><body>Multiple byte ranges not allowed</body></html>");     
 strcpy(phd->status_code,"406");
 return hdl_ok;
}

static void form_data_init(form_data *p) {
 p->isindex_query[0]=0;
 p->elements=0;
}

static void form_data_abandon_files(form_data *p) {
 form_element *e=p->elements;
 while(e!=0) {
     if(e->is_file) {
          e->is_file=false;
     }
     e=e->next;
 }
}

static void form_data_copy(form_data *dest, form_data *src) {
 dest->elements=src->elements;
 src->elements=0;
 strcpy(dest->isindex_query,src->isindex_query);
}

static void form_data_deinit(form_data *p) {
 form_element *e;
 while(p->elements!=0) {
     e=p->elements->next;
     if(p->elements->is_file) {
          unlink(p->elements->data);
     }
     checked_free(p->elements->name);
     checked_free(p->elements->data);
     checked_free(p->elements);
     p->elements=e;
 }
}

static void form_data_add(form_data *p, const char *name, const char *data, bool is_file) {
 form_element *e=(form_element*)checked_malloc(sizeof(form_element));
 if(e!=0) {
     e->name=checked_strdup(name);
     e->data=checked_strdup(data);
     e->is_file=is_file;
     e->next=p->elements;
     p->elements=e;
 }
}

static bool form_data_multipart_read(form_data *p, const char *boundary, PS_SOCKET so) {
 const char *dispo="Content-Disposition: form-data; name=";
 const char *fna="; filename=";
 char line[4096];
 char name[TOKSIZE]="";
 char value[TOKSIZE];
 bool is_file=false;
 FILE *file;
 typedef enum {
     parse_start,
     parse_part_header,
     parse_part_body,
 } parse_multi_t;
 parse_multi_t st=parse_start;


 while(PS_recv_line(so,line,sizeof(line)-1)) {
     if(memcmp(line,"--",2)==0 && memcmp(line+2,boundary,strlen(boundary))==0) {
          if(memcmp(line+strlen(boundary)+2,"--",2)==0) {
               if(st==parse_part_header || st==parse_start) {
                    return false;
               }
               if(is_file) {
                    fclose(file);
               }
               form_data_add(p,name,value,is_file);
               return true;
          }
          if(st==parse_part_header) {
               return false;
          }
          if(st==parse_part_body) {
               if(is_file) {
                    fclose(file);
               }
               form_data_add(p,name,value,is_file);
          }
          st=parse_part_header;
          name[0]=0;
          is_file=false;
          continue;
     }
     if(st==parse_part_header) {
          if(memicmp(line,dispo,strlen(dispo))==0) {
               char *p=line+strlen(dispo);
               if(*p=='"') {
                    char *cl_q=strchr(p+1,'"');
                    memcpy(name,p+1,cl_q-p-1);
                    name[cl_q-p-1]=0;
                    p=cl_q+1;
                    if(memicmp(p,fna,strlen(fna))==0) {
#ifdef NT
                         GetCurrentDirectory(TOKSIZE,value);
                         if(tmpnam(value+2)==0) { /* Retain current drive */
                              return false;
                         }
#else
                         if(tmpnam(value)==0) {
                              return false;
                         }
#endif
                         file=fopen(value,"w");
                         if(file==0) {
                              return false;
                         }
                         is_file=true;
                    }
               } else {
                    strcpy(name,p);
               }
          }
          if(*line==0) {
               st=parse_part_body;
          }
     } else if(st==parse_part_body) {
          if(is_file) {
               if(fprintf(file,"%s\n",line)<0) {
                    fclose(file);
                    return false;
               }
          } else {
               strcpy(value,line);
          }
     }
 }
 return false;
}

static bool form_data_url_encode(form_data *f, char *line) {
 char name[TOKSIZE];
 char value[TOKSIZE];
 form_element *e=f->elements;
 char *p=line;

 while(e!=0) {
     url_encode(e->name,name);
     url_encode(e->data,value);
     sprintf(p,"%s=%s%s",name,value,(e->next!=0?"&":""));
     p=strchr(p,0);
     e=e->next;
 }
 return true;
}

static bool form_data_url_decode(form_data *p, const char *line, bool isindex_allowed) {
 const char *s=line;
 const char *eq;
 char name[TOKSIZE];
 char value[TOKSIZE];

 if((eq=strchr(s,'='))!=0) {
     do {
          int count;
          if(eq-s>=TOKSIZE) {
               strncpy(name,s,TOKSIZE-1);
               name[TOKSIZE-1]=0;
          } else {
               strncpy(name,s,eq-s);
               name[eq-s]=0;
          }
          count=url_decode(eq+1,value);

          form_data_add(p,name,value,false);

          s=eq+1+count;
          if(*s=='&') {
               s++;
          }
     } while((eq=strchr(s,'='))!=0);
     return true;
 } else if(isindex_allowed) {
     strcpy(p->isindex_query,line);
     return false;
 }

 return false;
}

bool form_data_element(form_data *p, const char *name, char *value, bool *is_file) {
 form_element *e=p->elements;
 *value=0;
 while(e!=0) {
     if(stricmp(name,e->name)==0) {
          if(is_file!=0) {
               *is_file=e->is_file;
          }
          strcpy(value,e->data);
          return true;
     }
     e=e->next;
 }
 return false;
}

bool form_data_empty(form_data *p) {
 return p->elements==0;
}

static int url_decode(const char *urlenc, char *value) {
 int count=0;

 while(*urlenc!=0 && *urlenc!='&') {
     if(*urlenc == '+'){
          *value++ = ' ';
          count++;
          urlenc++;
     } else if(*urlenc=='%') {
          char c;
          int dig2, dig1=urlenc[1]-'0';
          if(dig1>15) {
               if(isupper(urlenc[1])) {
                    dig1=urlenc[1]-'A'+10;
               } else {
                    dig1=urlenc[1]-'a'+10;
               }
          }
          dig2=urlenc[2]-'0';
          if(dig2>15) {
               if(isupper(urlenc[2])) {
                    dig2=urlenc[2]-'A'+10;
               } else {
                    dig2=urlenc[2]-'a'+10;
               }
          }
          c=dig1*16+dig2;
          if(c == '\r') {
               c = ' ';
          }
          *value++=c;
          count+=3;
          urlenc+=3;
     } else {
          *value++=*urlenc++;
          count++;
     }
 }
 *value=0;
 return count;
}

static int url_encode(const char *plain, char *urlenc) {
 int count=0;

 while(*plain!=0) {
     if(isascii(*plain) && isalnum(*plain)) {
          *urlenc++=*plain++;
          count++;
     } else if(*plain==' ') {
          *urlenc++='+';
          plain++;
          count++;
     } else {
          sprintf(urlenc,"%%%02X",(int)(*plain));
          plain++;
          urlenc+=3;
          count+=3;
     }
 }
 *urlenc=0;
 return count;
}

static void base64decode_digit(const uint8 *base64, uint8 *plain) {
 uint8 bit6;
 char c=*base64;
 if(c>='A' && c<='Z') {
     bit6=c-'A';
 } else if(c>='a' && c<='z') {
     bit6=26+c-'a';
 } else if(c>='0' && c<='9') {
     bit6=52+c-'0';
 } else if(c=='+') {
     bit6=62;
 } else if(c=='/') {
     bit6=63;
 }
 *plain=bit6;
}

static void base64decode_quadruplet(const uint8 *base64, uint8 *p) {
 uint8 x[4];
 int chars_present;

 base64decode_digit(base64,x);
 base64decode_digit(base64+1,x+1);
 base64decode_digit(base64+2,x+2);
 base64decode_digit(base64+3,x+3);

 if(base64[3]=='=') {
     if(base64[2]=='=') {
          chars_present=1;
          p[1]=0;
     } else {
          chars_present=2;
          p[2]=0;
     }
 } else {
     chars_present=3;
     p[3]=0;
 }


 switch(chars_present) {
     case 3:
     p[2]='\xff' & (x[2]<<6);      /* First two bits */
     p[2]|=~'\xc0' & x[3];         /* Last six bits */

     case 2:
     p[1]='\xff' & (x[1]<<4);      /* First four bits */
     p[1]|='\x0f' & (x[2]>>2);    /* Last four bits */

     case 1:
     p[0]='\xff' & (x[0]<<2);      /* First six bits */
     p[0]|='\x03' & (x[1]>>4);    /* Last two bits */
 }
}

static void base64decode(const char *base64, char* plain, int base64len) {
 const uint8 *from=(uint8*)base64;
 uint8 *to=(uint8*)plain;
 for(;base64len!=0;base64len-=4,from+=4,to+=3) {
     base64decode_quadruplet(from,to);
 }
}

static bool same_host(struct sockaddr *p1, struct sockaddr *p2) {
 if(p1->sa_family!=p1->sa_family) {
     return false;
 }
 switch(p1->sa_family) {
     case AF_INET: {
          struct sockaddr_in *s1=(struct sockaddr_in*)p1;
          struct sockaddr_in *s2=(struct sockaddr_in*)p2;
          return memcmp(&(s1->sin_addr),&(s2->sin_addr),4)==0;
     } break;

#if defined(UNIX_LISTENER)
     case AF_UNIX: {
          struct sockaddr_un *s1=(struct sockaddr_un*)p1;
          struct sockaddr_un *s2=(struct sockaddr_un*)p2;
          return strcmp(s1->sun_path,s2->sun_path);
     } break;
#endif
 }

 return false;
}

void format_address(char *buffer, struct sockaddr *p, bool host_only) {
 switch(p->sa_family) {
     case AF_INET: {
          char port[20];
          struct sockaddr_in *s=(struct sockaddr_in*)p;
          unsigned char *q=(unsigned char*)&(s->sin_addr);
          if(host_only) {
               *port=0;
          } else {
               sprintf(port,":%d",s->sin_port);
          }
          sprintf(buffer,"%d.%d.%d.%d%s",q[0],q[1],q[2],q[3],port);
     } break;

#if defined(UNIX_LISTENER)
     case AF_UNIX: {
          struct sockaddr_un *s=(struct sockaddr_un*)p;
          strcpy(buffer,s->sun_path);
     } break;
#endif

     default: {
          strcpy(buffer,"Never-never land");
     } break;
 }
}

static char *rfc1123_month_names[] = {
 "Jan" , "Feb" , "Mar" , "Apr", "May" , "Jun" ,
 "Jul" , "Aug", "Sep" , "Oct" , "Nov" , "Dec"
};

static char *rfc1123_weekday_names[] = {
 "Sun" , "Mon" , "Tue" , "Wed", "Thu" , "Fri" , "Sat"
};

void format_rfc1123_time(char *buffer, time_t ti) {
 struct tm *ptm=gmtime(&ti);
 sprintf(buffer,"%s, %02d %s %4d %02d:%02d:%02d GMT",
     rfc1123_weekday_names[ptm->tm_wday], ptm->tm_mday,
     rfc1123_month_names[ptm->tm_mon], ptm->tm_year+1900,
     ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
}

void format_CommonLogFile_time(char *buffer, time_t ti) {
 struct tm *ptm=gmtime(&ti);
 sprintf(buffer,"%02d/%s/%4d:%02d:%02d:%02d +0000",
     ptm->tm_mday,
     rfc1123_month_names[ptm->tm_mon], ptm->tm_year+1900,
     ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
}

void format_base_URI(char *buffer, http_header *phd) {
 uint8 *i=(uint8*)&my_IP_address;
 char s_port[10];
 const char *protocol = (is_secure?"https":"http");
 if(http_port==(is_secure?443:80)) {
     s_port[0]=0;
 } else {
     sprintf(s_port,":%d",http_port);
 }

 if(phd->host!=0) {
     sprintf(buffer,"%s://%s",protocol,phd->host);
 } else {
     sprintf(buffer,"%s://%d.%d.%d.%d%s",protocol,i[0],i[1],i[2],i[3],s_port);
 }
}

/* Random numbers */
static uint32 seed;

static random_seed(uint32 s) {
 seed=s;
}

static uint32 random_next() {
 static sint32 a=16807;
 static sint32 m=2147483647L;
 static sint32 q;
 static sint32 r;
 q=m/a;
 r=m%a;
 seed=a*(seed-(seed/q)*q)-(seed/q)*r;
 if(seed<0) seed+=m;
 return seed;
}

static void format_time(char *buffer, time_t ti) {
 struct tm *ptm=localtime(&ti);
 sprintf(buffer,"%04d-%02d-%02d %02d:%02d:%02d",
     ptm->tm_year+1900,ptm->tm_mon+1,ptm->tm_mday,ptm->tm_hour,ptm->tm_min,ptm->tm_sec);
}

void htmlize(const char *src, char *tgt) {
 while(*src!=0) {
     switch(*src) {
          case '<': strcpy(tgt,"&lt;"); tgt+=4; break;
          case '>': strcpy(tgt,"&gt;"); tgt+=4; break;
          case '&': strcpy(tgt,"&amp;"); tgt+=5; break;
          case '"': strcpy(tgt,"&quot;"); tgt+=6; break;
          default: *tgt++=*src; break;
     }
     src++;
 }
 *tgt=0;
}
