/* http.h */

#ifndef _HTTP_H
#define _HTTP_H

#include <time.h>
#include "header.h"
#include "machine.h"
#include "ps_sock.h"

enum {
 TOKSIZE=8192
};

#define HTTP_BUFFER_SZ (100*1024)

typedef enum {
 auth_none,
 auth_failed_passwd,
 auth_failed_cert,
 auth_unsup,
 auth_ok
} auth_outcome;

typedef enum {
 http_unknown,
 http_get,
 http_head,
 http_options,
 http_post,
 http_put,
 http_delete,
 http_trace
} http_req_t;

typedef enum {
 hdl_ok,       /* Ok, handled by this handler */
 hdl_newses,   /* Will be handled by this handler but needs session */
 hdl_error,    /* Error while processing */
 hdl_sep,      /* Somebody Else's Problem (not handled) */
 hdl_meth,     /* Method not supported */
 hdl_down,     /* Shutdown the server */
 hdl_defer,    /* Execution is defered to some child */
 hdl_defer_deinit, /* Execution is defered to some child process, deinit*/
 hdl_auth,     /* Not authorized */
 hdl_redirect  /* Redirect to other handler */
} hdl_t;

typedef struct form_element {
 char *name;
 bool is_file;
 char *data;
 uint32 length;
 struct form_element *next;
} form_element;

typedef struct {
 char isindex_query[TOKSIZE];
 form_element *elements;
} form_data;

typedef struct {
 char content_type[300];
 char referer[TOKSIZE];
 char user_agent[TOKSIZE];
 char cookie[TOKSIZE];
 char accept[TOKSIZE];
 char http_version[5];

 http_req_t orig_meth;
 http_req_t meth;
 char orig_request_URI[TOKSIZE];
 char request_URI[TOKSIZE];
 char *querystring;
 char status_code[4];
 unsigned long content_length;
 char path_info[TOKSIZE];
 char host[256];
 auth_outcome auth;
 sint16 uid;
 char user[50];
 char passwd[50];
 bool full_messages;
 uint32 sessionID;
 uint32 referer_sessionID;
 struct sockaddr peer_addr;
 form_data other_headers;
 form_data form;
 uint32 sent_bytes;
} http_header;

typedef struct {
 PS_SOCKET so;
 const char *realm;
 char *buffer;
 http_header* phd;
} handler_args;

/* realm_data::destroy deconstructs (but frees not) realm_data::data */
typedef struct realm_data {
 void *data;
 void (*destroy)(void*);
} realm_data;

typedef struct realm_link {
 char *realm;
 realm_data *object;
 struct realm_link *next;
} realm_link;

typedef struct {
 realm_link *realms;
} session_data;

/* A handler func is supposed to provide either all content (hdl_ok)
   or no content at all (rest of hdl_*) */

typedef hdl_t (*handler_func)(PS_SOCKET so, const char *realm, char *buffer, http_header*);
typedef hdl_t (*mime_type_handler_func)(PS_SOCKET so, char *buffer, http_header* phd, const char *fname);
typedef auth_outcome (*auth_func)(const char *uname, const char *pass, const char* host, sint16 *puid);
typedef bool (*email_to_user_func)(char username[], sint16 *puid, const char email[]);
typedef void (*shutdown_func)();
typedef bool (*critical_section_func)();

typedef struct {
 char *prefix;
 handler_func func;
} handler_s;

typedef struct {
 char *mime_type;
 mime_type_handler_func func;
} mime_type_handler_s;

/* Initialization/Configuration */
void http_server_init(char *hostname, uint32 hostaddr, int port, bool is_secure);
void http_server_set_critical_section_primitives(
     critical_section_func enter,
     critical_section_func leave);
void http_server_set_handlers(handler_s *h, int h_nb);
void http_server_set_mime_type_handlers(mime_type_handler_s *m, int m_nb);
void http_server_set_auth_func(auth_func authenticate);
void http_server_set_shutdown_func(shutdown_func shutdown);
void http_server_set_email_to_user_func(email_to_user_func email2user);
void http_server_set_httplogfile(const char *s);
void http_server_set_httprootdir(const char *s);
void http_server_set_httpshareddir(const char *s);
void http_server_set_execext(const char *s);
void http_server_set_binpath(const char *s);
void http_server_set_mimetypesfile(const char *s);
void http_server_set_httpsessiontimeout(int i);
void http_server_set_sslauthentication(int i);
void http_server_add_to_product_token(const char *s);
void http_server_additional_env_vars(char **pp, int nb);

/* Exported standard handler */
hdl_t file_handler(PS_SOCKET so, const char *realm, char *buffer, http_header* phd);
hdl_t file_handler_nothreads(PS_SOCKET so, const char *realm, char *buffer, http_header* phd);

/* Service routine */
void http_request(PS_SOCKET so, struct sockaddr *peer, char *buffer);

/* Session data */
realm_data *realm_data_make(void* data, void (*destroy)(void*));
void session_initialize();
bool session_set_realm_data(uint32 sessionID, const char *realm, realm_data *data);
realm_data *session_get_realm_data(uint32 sessionID, const char *realm);
void session_close(uint32 sessionID);
void session_html_dump(PS_SOCKET so, const char *delete_session_URL_prefix);

/* Form data */
bool handler_read_form_data(http_header *phd, PS_SOCKET so);
bool form_data_empty(form_data *p);
bool form_data_element(form_data *, const char *name, char *value, bool *is_file);

/* Restricted content negotiation */
bool http_demands_wap(http_header *phd);

/* Exported helper functions */
void http_send_file_data(PS_SOCKET so, const char *file, char *buffer);
void http_send_file_data_range(PS_SOCKET so, const char *file, char *buffer,
     uint32 offset_s, uint32 offset_e);
void http_redirect(PS_SOCKET so, http_header *phd, const char *fmt, ...);

hdl_t http_produce_external_content(
     PS_SOCKET so,
     char *buffer,
     http_header *,
     const char *cmdline);
hdl_t http_async_produce_external_content(
     PS_SOCKET so,
     char *buffer,
     http_header *,
     const char *cmdline);

const char *http_server_product_token();     
/* Exported utilities */     
void format_base_URI(char *buffer, http_header *);
const char *http_req_t_to_string(http_req_t meth);
void format_address(char *buffer, struct sockaddr *p, bool host_only);
void format_rfc1123_time(char *buffer, time_t);
void format_CommonLogFile_time(char *buffer, time_t);
void htmlize(const char *src, char *tgt);


#endif
