/* like.c */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "chkalloc.h"
#include "like.h"

/* Warning: char and uni_char versions should be kept in sync! */

enum { wildcard=1, any=2 };
#define MAX_NFA_STATES 300
#define MAX_DFA_STATES 3000

typedef struct {
 uint8 flags;
 int symbol;
 int state; /* Not used for NFA */
} flagged_transition;

typedef struct {
 int num;
 int accepting_state;
 flagged_transition transitions[MAX_NFA_STATES];
} nfa;

typedef struct {
 uint8 bitvec[MAX_NFA_STATES/8];
 bool accepting;
 int num_transitions;
 flagged_transition transitions[10];
} temp_dfa_state_set;

typedef struct {
 int num;
 int processed;
 temp_dfa_state_set state_sets[MAX_DFA_STATES];
} temp_dfa;

typedef struct {
 bool accepting;
 int num_transitions;
 flagged_transition *transitions;
} dfa_state_set;

typedef struct dfa {
 int num;
 dfa_state_set state_sets[1]; /* To provide alignment */
} dfa;

static void bitvec_clear(uint8 *s);
static void bitvec_set(uint8 *s, int i, int x);
static int bitvec_get(const uint8 *s, int i);

static bool construct_nfa(nfa *p, const char *pattern, char escape);
static bool uni_construct_nfa(nfa *p, const uni_char *pattern, uni_char escape);
static bool construct_temp_dfa(const nfa *p, temp_dfa *d);

static void add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(
     const nfa *p, temp_dfa *d, int nfa_state, int dfa_state);
static bool same_states(const temp_dfa *d, int i, int j);
static bool string_recognized_by_temp_dfa(const temp_dfa *d, const char *string);
static bool step_temp_dfa(const temp_dfa *d, char symbol, int *pstate);
static bool step_dfa(const dfa *d, char symbol, int *pstate);
static bool uni_string_recognized_by_temp_dfa(const temp_dfa *d, const uni_char *string);
static bool uni_step_temp_dfa(const temp_dfa *d, uni_char symbol, int *pstate);
static bool uni_step_dfa(const dfa *d, uni_char symbol, int *pstate);
static const dfa *temp_dfa_to_dfa(const temp_dfa *d);

#if defined(DEBUG_LIKE)
static void output_nfa(const nfa *p);     
static void output_temp_dfa(const temp_dfa *d);
static void output_dfa(const dfa *d);
#endif

void free_dfa(const void *d) {
 checked_free((void*)d);
}

bool string_like(const char *string, const char *pattern, char escape) {
 nfa a;
 temp_dfa *ptd;
 bool ret;
 if(!construct_nfa(&a,pattern,escape)) {
     return false;
 }
#if defined(DEBUG_LIKE)
 output_nfa(&a);
#endif
 ptd=checked_malloc(sizeof(temp_dfa));
 if(ptd==0) {
     checked_free(ptd);
     return false;
 }
 if(!construct_temp_dfa(&a,ptd)) {
     return false;
 }
#if defined(DEBUG_LIKE)
 output_temp_dfa(ptd);
#endif
 ret=string_recognized_by_temp_dfa(ptd,string);
 checked_free(ptd);
 return ret;
}

bool uni_string_like(const uni_char *string, const uni_char *pattern, uni_char escape) {
 nfa a;
 temp_dfa *ptd;
 bool ret;
 if(!uni_construct_nfa(&a,pattern,escape)) {
     return false;
 }
#if defined(DEBUG_LIKE)
 output_nfa(&a);
#endif
 ptd=checked_malloc(sizeof(temp_dfa));
 if(ptd==0) {
     return false;
 }
 if(!construct_temp_dfa(&a,ptd)) {
     checked_free(ptd);
     return false;
 }
#if defined(DEBUG_LIKE)
 output_temp_dfa(ptd);
#endif
 ret=uni_string_recognized_by_temp_dfa(ptd,string);
 checked_free(ptd);
 return ret;
}

const void *pattern_to_dfa(const char *pattern, char escape) {
 nfa a;
 temp_dfa *ptd;
 const void *ret;
 if(!construct_nfa(&a,pattern,escape)) {
     return 0;
 }
#if defined(DEBUG_LIKE)
 output_nfa(&a);
#endif
 ptd=checked_malloc(sizeof(temp_dfa));
 if(ptd==0) {
     return 0;
 }
 if(!construct_temp_dfa(&a,ptd)) {
     checked_free(ptd);
     return 0;
 }
#if defined(DEBUG_LIKE)
 output_temp_dfa(ptd);
#endif
 ret=temp_dfa_to_dfa(ptd);
 checked_free(ptd);
 return ret;
}

const void *uni_pattern_to_dfa(const uni_char *pattern, uni_char escape) {
 nfa a;
 temp_dfa *ptd;
 const void *ret;
 if(!uni_construct_nfa(&a,pattern,escape)) {
     return 0;
 }
#if defined(DEBUG_LIKE)
 output_nfa(&a);
#endif
 ptd=checked_malloc(sizeof(temp_dfa));
 if(ptd==0) {
     return 0;
 }
 if(!construct_temp_dfa(&a,ptd)) {
     return 0;
 }
#if defined(DEBUG_LIKE)
 output_temp_dfa(ptd);
#endif
 ret=temp_dfa_to_dfa(ptd);
 checked_free(ptd);
 return ret;
}

/* bit vector utilities */
static void bitvec_clear(uint8 *s) {
 int i;
 for(i=0; i<MAX_NFA_STATES/8; i++) {
     s[i]=0;
 }
}

static void bitvec_set(uint8 *s, int i, int x) {
 uint8 *p=s+i/8;
 int b=i%8;
 if(p) {
     if(x) {
          (*p)|=(1<<b);
     } else {
          (*p)&=~(1<<b);
     }
 }
}

static int bitvec_get(const uint8 *s, int i) {
 const uint8 *p=s+i/8;
 int b=i%8;
 return ((*p)&(1<<b))!=0; 
}

/* Helper graph routines */
static bool construct_nfa(nfa *p, const char *pattern, char escape) {
 int i=0;
 int is;
 for(is=0; is<MAX_NFA_STATES; is++) {
     p->transitions[is].flags=0;
     p->transitions[is].symbol=0;
 }
 is=0;
 for(i=0; pattern[i]!=0; i++) {
     if(is==MAX_NFA_STATES) {
          return false;
     }
     switch(pattern[i]) {
          case '%':
               if(is!=0) {
                    /* Collapse juxtaposed wildcards */
                    if(p->transitions[is-1].flags&wildcard
                    && p->transitions[is-1].symbol==0) { /* %% -> % */
                         break;
                    }
               }                         
               p->transitions[is].flags|=wildcard;
               break;
          case '_':
               p->transitions[is++].flags|=any;
               break;
          default:
               if(pattern[i]==escape) {
                    i++;
                    p->transitions[is++].symbol=pattern[i];
                    break;
               }
               p->transitions[is++].symbol=pattern[i];
               break;
     }
 }
 if(p->transitions[is].flags&wildcard) {
     p->accepting_state=is;
     is++;
 } else {
     p->accepting_state=is;
 }
 p->num=is;
 return true;
}

static bool uni_construct_nfa(nfa *p, const uni_char *pattern, uni_char escape) {
 int i=0;
 int is;
 for(is=0; is<MAX_NFA_STATES; is++) {
     p->transitions[is].flags=0;
     p->transitions[is].symbol=0;
 }
 is=0;
 for(i=0; pattern[i]!=0; i++) {
     if(is==MAX_NFA_STATES) {
          return false;
     }
     switch(pattern[i]) {
          case '%':
               if(is!=0) {
                    /* Collapse juxtaposed wildcards */
                    if(p->transitions[is-1].flags&wildcard) { /* %% -> % */
                         break;
                    }
                    if(p->transitions[is-1].flags&any) {
                         p->transitions[is-1].flags=wildcard; /* _% -> % */
                         break;
                    }     
               }                         
               p->transitions[is].flags|=wildcard;
               break;
          case '_':
               if(is!=0) {
                    /* Collapse juxtaposed wildcards */
                    if(p->transitions[is-1].flags&wildcard) { /* %_ -> % */
                         break;
                    }
               }
               p->transitions[is++].flags|=any;
               break;
          default:
               if(pattern[i]==escape) {
                    i++;
                    p->transitions[is++].symbol=pattern[i];
                    break;
               }
               p->transitions[is++].symbol=pattern[i];
               break;
     }
 }
 if(p->transitions[is].flags&wildcard) {
     p->accepting_state=is;
     is++;
 } else {
     p->accepting_state=is;
 }
 p->num=is;
 return true;
}

static bool construct_temp_dfa(const nfa *p, temp_dfa *d) {
 int i,j;
 
 /* Initialize the temp_dfa */
 d->num=d->processed=0;
 for(i=0; i<MAX_DFA_STATES; i++) {
     bitvec_clear(d->state_sets[i].bitvec);
     d->state_sets[i].num_transitions=0;
     d->state_sets[i].accepting=0;
     for(j=0; j<10; j++) {
          d->state_sets[i].transitions[j].flags=0;
          d->state_sets[i].transitions[j].state=0;
          d->state_sets[i].transitions[j].symbol=0;
     }
 }
 
 /* Initially, start with T0=epsilon-closure(state-0) */
 d->num=1;
 add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(p,d,0,0);

 /* Main loop, processing unmarked states */
 while(d->processed<d->num) {
     int i;
     bool any_needed=false;
     int ix=d->processed;
     /* We must record transitions on every input symbol. In practice,
        we limit ourselves to symbols referenced in the NFA and the catch-all
        "any", where applicable */

     /* First pass, gather all symbols required for transition table, plus
        the special "any" symbol */
     for(i=0; i<p->num; i++) {
          /* For all states in DFA state */
          if(bitvec_get(d->state_sets[ix].bitvec,i)) {
               if(p->transitions[i].flags&wildcard || 
                  p->transitions[i].flags&any) {
                    /* If the transition is on "any", just record the fact,
                       because we'll add the corresponding transition to the
                       end of the transition table */
                    any_needed=true;
               }
               if(p->transitions[i].symbol!=0) {
                    /* If the transition is on a normal symbol, add it to the
                    transition table if not already there */
                    char symbol=p->transitions[i].symbol;
                    for(j=0; j<d->state_sets[ix].num_transitions; j++) {
                         if(d->state_sets[ix].transitions[j].symbol==symbol) {
                              break;
                         }
                    }
                    if(j==d->state_sets[ix].num_transitions) {
                         d->state_sets[ix].transitions[j].symbol=symbol;
                         d->state_sets[ix].num_transitions++;
                    }
               }   
          }
     }
     if(any_needed) {
          d->state_sets[ix].transitions[d->state_sets[ix].num_transitions++].flags|=any;
     }

     /* Second pass, accumulate epsilon-closure(move(state-i,symbol)) for all
        the symbols required */
     for(j=0; j<d->state_sets[ix].num_transitions; j++) {
          bool transition_on_any=false;
          char symbol;
          if(d->state_sets[ix].transitions[j].flags&any) {
               transition_on_any=true;
          } else {
               symbol=d->state_sets[ix].transitions[j].symbol;
          }                         
          for(i=0; i<p->num; i++) {
               /* For all states in DFA state */
               if(bitvec_get(d->state_sets[ix].bitvec,i)) {
                    if(p->transitions[i].flags&wildcard) {
                         add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(p,d,i,d->num);
                    }                         
                    if(transition_on_any && p->transitions[i].flags&any) {
                         add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(p,d,i+1,d->num);                         
                    } else if(!transition_on_any && p->transitions[i].symbol==symbol) {
                         add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(p,d,i+1,d->num);
                    }
               }
          }
          /* See if this state-set has appeared before */
          for(i=0; i<d->num; i++) {
               if(same_states(d,i,d->num)) {
                    bitvec_clear(d->state_sets[d->num].bitvec);
                    break;
               }
          }
          d->state_sets[ix].transitions[j].state=i;
          if(i==d->num) {
               d->num++;
          }
     }

     /* See if ANY transition subsumes transitions on symbols and
        asimilate them */
     if(d->state_sets[ix].num_transitions>0 
     && d->state_sets[ix].transitions[d->state_sets[ix].num_transitions-1].flags&any) {
          int state=d->state_sets[ix].transitions[d->state_sets[ix].num_transitions-1].state;
          for(j=0; j<d->state_sets[ix].num_transitions-1; j++) {
               if(d->state_sets[ix].transitions[j].state==state) {
                    int k;
                    for(k=j; k<d->state_sets[ix].num_transitions-1; k++) {
                         d->state_sets[ix].transitions[k]=d->state_sets[ix].transitions[k+1];
                    }
                    d->state_sets[ix].num_transitions--;
               }
          }
     }

     /* Mark state as accepting or not */
     if(bitvec_get(d->state_sets[ix].bitvec,p->accepting_state)) {
          d->state_sets[ix].accepting=true;
     }

     /* Proceed to the next */
     d->processed++;
 }

 return true;
}

static const dfa *temp_dfa_to_dfa(const temp_dfa *d) {
 dfa *ret;
 int i;
 flagged_transition *tr;
 int boundary;
 int tot_size;
 int tot_num_transitions;

 tot_num_transitions=0;
 for(i=0; i<d->num; i++) {
     tot_num_transitions+=d->state_sets[i].num_transitions;
 }
 
 boundary=sizeof(dfa)+(d->num-1)*sizeof(dfa_state_set);
 boundary+=boundary%sizeof(flagged_transition);

 tot_size=boundary+tot_num_transitions*sizeof(flagged_transition);

 ret=(dfa*)checked_malloc(tot_size);
 if(ret==0) {
     return 0;
 }

 tr=(flagged_transition*)((char*)ret+boundary);

 ret->num=d->num;

 for(i=0; i<d->num; i++) {
     ret->state_sets[i].num_transitions=d->state_sets[i].num_transitions;
     ret->state_sets[i].accepting=d->state_sets[i].accepting;
     ret->state_sets[i].transitions=tr;
     memcpy(tr,d->state_sets[i].transitions,
          ret->state_sets[i].num_transitions*sizeof(flagged_transition));
     tr+=ret->state_sets[i].num_transitions;
 }
 
 return ret;
}

static void add_epsilon_closure_of_nfa_state_to_temp_dfa_state_set(
     const nfa *p, temp_dfa *d, int nfa_state, int dfa_state) {
 bitvec_set(d->state_sets[dfa_state].bitvec,nfa_state,1);
}     

static bool same_states(const temp_dfa *d, int i, int j) {
 return memcmp(d->state_sets[i].bitvec,d->state_sets[j].bitvec,MAX_NFA_STATES/8)==0;
}

bool string_recognized_by_dfa(const void *v, const char *string) {
 const dfa *d=(const dfa *)v;
 int state=0; /* Initial state */
 while(*string!=0) {
     if(!step_dfa(d,*string,&state)) {
          return false;
     }
     string++;
 }
 return d->state_sets[state].accepting;
}

static bool string_recognized_by_temp_dfa(const temp_dfa *d, const char *string) {
 int state=0; /* Initial state */
 while(*string!=0) {
     if(!step_temp_dfa(d,*string,&state)) {
          return false;
     }
     string++;
 }
 return d->state_sets[state].accepting;
}

static bool step_temp_dfa(const temp_dfa *d, char symbol, int *pstate) {
 int ix=*pstate;
 int i;
 for(i=0; i<d->state_sets[ix].num_transitions; i++) {
     if(d->state_sets[ix].transitions[i].flags&any
     || d->state_sets[ix].transitions[i].symbol==symbol) {
          *pstate=d->state_sets[ix].transitions[i].state;
          return true;
     }
 }
 return false;
}

static bool step_dfa(const dfa *d, char symbol, int *pstate) {
 int ix=*pstate;
 int i;
 for(i=0; i<d->state_sets[ix].num_transitions; i++) {
     if(d->state_sets[ix].transitions[i].flags&any
     || d->state_sets[ix].transitions[i].symbol==symbol) {
          *pstate=d->state_sets[ix].transitions[i].state;
          return true;
     }
 }
 return false;
}

bool uni_string_recognized_by_dfa(const void *v, const uni_char *string) {
 const dfa *d=(const dfa *)v;
 int state=0; /* Initial state */
 while(*string!=0) {
     if(!uni_step_dfa(d,*string,&state)) {
          return false;
     }
     string++;
 }
 return d->state_sets[state].accepting;
}

static bool uni_string_recognized_by_temp_dfa(const temp_dfa *d, const uni_char *string) {
 int state=0; /* Initial state */
 while(*string!=0) {
     if(!uni_step_temp_dfa(d,*string,&state)) {
          return false;
     }
     string++;
 }
 return d->state_sets[state].accepting;
}

static bool uni_step_temp_dfa(const temp_dfa *d, uni_char symbol, int *pstate) {
 int ix=*pstate;
 int i;
 for(i=0; i<d->state_sets[ix].num_transitions; i++) {
     if(d->state_sets[ix].transitions[i].flags&any
     || d->state_sets[ix].transitions[i].symbol==symbol) {
          *pstate=d->state_sets[ix].transitions[i].state;
          return true;
     }
 }
 return false;
}

static bool uni_step_dfa(const dfa *d, uni_char symbol, int *pstate) {
 int ix=*pstate;
 int i;
 for(i=0; i<d->state_sets[ix].num_transitions; i++) {
     if(d->state_sets[ix].transitions[i].flags&any
     || d->state_sets[ix].transitions[i].symbol==symbol) {
          *pstate=d->state_sets[ix].transitions[i].state;
          return true;
     }
 }
 return false;
}

#if defined(DEBUG_LIKE)
void main(int argc, char **argv) {
 const dfa *d;
 if(argc!=3) {
     printf("Usage: like <pattern> <string>\n");
     return;
 }
 printf("Step1: string_like : ");
 if(!string_like(argv[2],argv[1],0)) {
     printf("Does not match!\n");
 } else {
     printf("Matches!\n");
 }
 printf("Step2: string_to_dfa\n");
 d=pattern_to_dfa(argv[1],0);
 if(d==0) {
     printf("Not produced!\n");
     return;
 }
 output_dfa(d);
 if(!string_recognized_by_dfa(d,argv[2])) {
     printf("Does not match!\n");
 } else {
     printf("Matches!\n");
 }
 checked_free(d);
}

static void output_temp_dfa(const temp_dfa *d) {
 int i, j;
 for(i=0; i<d->num; i++) {
     printf("%d {",i);
     for(j=0; j<MAX_NFA_STATES; j++) {
          if(bitvec_get(d->state_sets[i].bitvec,j)) {
               printf("%d ",j);
          }
     }
     printf("} ");
     for(j=0; j<d->state_sets[i].num_transitions; j++) {
          if(d->state_sets[i].transitions[j].flags&any) {
               printf("-ANY->%d ",d->state_sets[i].transitions[j].state);
          } else {
               printf("-%c->%d ",d->state_sets[i].transitions[j].symbol,
                    d->state_sets[i].transitions[j].state);
          }
     }
     if(d->state_sets[i].accepting) {
          printf("(Accepting)");
     }
     printf("\n");
 }
}

static void output_dfa(const dfa *d) {
 int i, j;
 for(i=0; i<d->num; i++) {
     printf("%d ",i);
     for(j=0; j<d->state_sets[i].num_transitions; j++) {
          if(d->state_sets[i].transitions[j].flags&any) {
               printf("-ANY->%d ",d->state_sets[i].transitions[j].state);
          } else {
               printf("-%c->%d ",d->state_sets[i].transitions[j].symbol,
                    d->state_sets[i].transitions[j].state);
          }
     }
     if(d->state_sets[i].accepting) {
          printf("(Accepting)");
     }
     printf("\n");
 }
}

static void output_nfa(const nfa *p) {
 int i;
 for(i=0; i<p->num; i++) {
     printf("\n%d: ");
     if(p->transitions[i].flags&wildcard) printf("(wildcard) ");
     if(p->transitions[i].flags&any) printf("transition on ANY ");
     if(p->transitions[i].symbol!=0) printf("transition on %c",p->transitions[i].symbol);
 }
 printf("\nAccepting: %d\n",p->accepting_state);
}

#endif

