/* OPENEXEC.C */
#include <stdarg.h>
#include <stdlib.h>

#if defined(UNIX)
#include <sys/param.h>
#include <unistd.h>
#include <fcntl.h>

# if !defined(NOFILE)
#  include <stdio.h>
#  define NOFILE _NFILE
# endif
#else /* NT */
#include <direct.h>
#endif

#include "err_sys.h"
#include "altera.h"
#include "openexec.h"

static void chdir_to_path(const char *cmdline) {
 const char *last=cmdline;
 const char *orig=cmdline;
 for(; *cmdline!=0 && *cmdline!=' '; cmdline++) {
     if(*cmdline=='/'
#if defined(NT)
     || *cmdline=='\\'
#endif
     ) {
          last=cmdline;
     }
 }
 if(last!=orig) {
     char path[1025];
     /* Trailing delimiter is not portable */
     strncpy(path,orig,last-orig);
     path[last-orig]=0;
     chdir(path);
 }
}

#if defined(UNIX)
static void close_other_fds() {
 int fd;
 for(fd=3; fd<NOFILE; fd++) {
     close(fd);
 }
 fd=open("/dev/null", O_WRONLY);
 dup2( fd, 2);
 close(fd);
}

int openexecl(const char *cmd, const char *arg0, ...) {
 const char *args[100];
 int i=0;

 va_list marker;
 va_start(marker,arg0);

 args[i++]=cmd;
 args[i++]=arg0;
 for(;;) {
     args[i]=va_arg(marker,const char *);
     if(args[i]==0) {
          break;
     }
     i++; 
 }
 va_end(marker);
 
 return openexecv(cmd,args);
}

int openexecv(const char *cmd, const char * const *args) {
 int fd[2];
 int child_pid;

 if(pipe(fd)<0) {
     return -1;
 }

 child_pid=fork();
 if(child_pid<0) {
     close(fd[0]);
     close(fd[1]);
     return -1;
 } else if(child_pid!=0) {
     close(fd[1]);
     return fd[0];
 } else {
     close(fd[0]);
     close(1);
     dup(fd[1]);
     close(fd[1]);
     close_other_fds();
     chdir_to_path(cmd);
     execv(cmd,args);
     exit(1);
 }
}
#endif

int pipexecle(PIPE *fd2, const char *cmd, const char **envp, const char *arg0, ...) {
 const char *args[100];
 int i=0;

 va_list marker;
 va_start(marker,arg0);

 args[i++]=arg0;
 for(;;) {
     args[i]=va_arg(marker,const char *);
     if(args[i]==0) {
          break;
     }
     i++; 
 }
 va_end(marker);

 return pipexecve(fd2,cmd,envp,args);
}

#if defined(UNIX)

int pipexecve(PIPE *fd, const char *cmd, const char **envp, const char * const *args) {
 int child_pid;
 int fd1[2];
 int fd2[2];
/*
      fd1[1]  >--------> fd1[0]
  sqlcore                        AGI
      fd2[0]  <--------< fd2[1]
*/

 if(pipe(fd1)<0) {
     return 0;
 }
 if(pipe(fd2)<0) {
     close(fd1[0]);
     close(fd1[1]);
     return 0;
 }

 child_pid=fork();
 if(child_pid<0) {
     close(fd1[0]);
     close(fd1[1]);
     close(fd2[0]);
     close(fd2[1]);
     return 0;
 } else if(child_pid!=0) {
     close(fd1[0]);
     close(fd2[1]);
     fd[0]=fd2[0];
     fd[1]=fd1[1];
     return 1;
 } else {
     int ret;
     close(fd1[1]);
     close(fd2[0]);
     close(0);
     dup(fd1[0]);
     close(fd1[0]);
     close(1);
     dup(fd2[1]);
     close(fd2[1]);
     close_other_fds();
     chdir_to_path(cmd);
     if(envp==0) {
          ret=execve(cmd,args,envp);
     } else {
          ret=execve(cmd,args,envp);
     }
     if(ret<0) {
          warn_sys("pipexecv error");
     }
     exit(1);
 }
}

#else

#if 0 && defined(SERVICE)
int pipexecve(PIPE *fd, const char *cmd, const char **envp, const char * const *args) {
 PIPE fd1[2];
 PIPE fd2[2];
 SECURITY_ATTRIBUTES sat;
 PROCESS_INFORMATION pi;
 STARTUPINFO si;
 int ret;
 char cmdline1[1024];
 char *cmdline=cmdline1+1;
 char dir[1025];
 PIPE dupWr;
 PIPE hin=GetStdHandle(STD_INPUT_HANDLE);
 PIPE hout=GetStdHandle(STD_OUTPUT_HANDLE);
 char envblock[ENVSIZE]="";
 char *envblockp=0;

 cmdline1[0]=0;
 for(; *args!=0; args++) {
     strcat(cmdline1," \"");
     strcat(cmdline1,*args);
     strcat(cmdline1,"\"");
 }

 if(envp!=0) {
     envblock[0]=0;
     envblockp=envblock;
     for(; *envp!=0; envp++) {
          int len=strlen(*envp);
          memcpy(envblockp,*envp,len+1);
          envblockp+=len+1;
     }
     *envblockp=0;
     envblockp=envblock;
 }

/*
      fd1[1]  >--------> fd1[0]
  sqlcore                        AGI
      fd2[0]  <--------< fd2[1]
*/

 if(!DIR_getDirectoryWithoutSep(cmdline, dir)) {
     return 0;
 }

 sat.nLength=sizeof(sat);
 sat.lpSecurityDescriptor=0;
 sat.bInheritHandle=TRUE;
 if(CreatePipe(fd1,fd1+1,&sat,0)<0) {
     return 0;
 }

 sat.nLength=sizeof(sat);
 sat.lpSecurityDescriptor=0;
 sat.bInheritHandle=TRUE;
 if(CreatePipe(fd2,fd2+1,&sat,0)<0) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     return 0;
 }

 if(!SetStdHandle(STD_INPUT_HANDLE,fd1[0])) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }

 if(!SetStdHandle(STD_OUTPUT_HANDLE,fd2[1])) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }

 si.cb=sizeof(si);
 si.lpReserved=si.lpReserved2=0;
 si.cbReserved2=0;
 si.lpDesktop=0;
 si.lpTitle=0;
 si.dwFlags=STARTF_USESHOWWINDOW;
 si.wShowWindow=SW_HIDE;

 if(!DuplicateHandle(GetCurrentProcess(),fd1[1],GetCurrentProcess(),
     &dupWr,0,FALSE,DUPLICATE_SAME_ACCESS)) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }
 pipe_close(fd1[1]);
 fd1[1]=dupWr;

 ret=CreateProcess(cmd,cmdline,0,0,TRUE,0,envblockp,dir,&si,&pi);

 pipe_close(fd1[0]); /*ommited for some reason from example*/
 pipe_close(fd2[1]);

 if(ret) {
     fd[0]=fd2[0];
     fd[1]=fd1[1];
 } else {
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     warn_sys("CreateProcess %ld",GetLastError());
 }

 return ret;
}
#else
/*In order to get the effect of pipe() in Windows:
  1) Ignore the example in the section on "Pipe". It only works with
     console applications
  2) Fill the appropriate hStd* members in the STARTUPINFO */
int pipexecve(PIPE *fd, const char *cmd, const char **envp, const char * const *args) {
 PIPE fd1[2];
 PIPE fd2[2];
 SECURITY_ATTRIBUTES sat;
 PROCESS_INFORMATION pi;
 STARTUPINFO si;
 int ret;
 char cmdline1[1025];
 char *cmdline=cmdline1+1;
 char dir[1025];
 PIPE dupWr;
 PIPE hin=GetStdHandle(STD_INPUT_HANDLE);
 PIPE hout=GetStdHandle(STD_OUTPUT_HANDLE);
 char envblock[ENVSIZE]="";
 char *envblockp=0;

 cmdline1[0]=0;
 for(; *args!=0; args++) {
     strcat(cmdline1," \"");
     strcat(cmdline1,*args);
     strcat(cmdline1,"\"");
 }

 if(envp!=0) {
     envblock[0]=0;
     envblockp=envblock;
     for(; *envp!=0; envp++) {
          int len=strlen(*envp);
          memcpy(envblockp,*envp,len+1);
          envblockp+=len+1;
     }
     *envblockp=0;
     envblockp=envblock;
 }

/*
      fd1[1]  >--------> fd1[0]
  sqlcore                        AGI
      fd2[0]  <--------< fd2[1]
*/

 if(!DIR_getDirectoryWithoutSep(cmd, dir)) {
     return 0;
 }

 sat.nLength=sizeof(sat);
 sat.lpSecurityDescriptor=0;
 sat.bInheritHandle=TRUE;
 if(CreatePipe(fd1,fd1+1,&sat,0)<0) {
     return 0;
 }

 sat.nLength=sizeof(sat);
 sat.lpSecurityDescriptor=0;
 sat.bInheritHandle=TRUE;
 if(CreatePipe(fd2,fd2+1,&sat,0)<0) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     return 0;
 }

 if(!SetStdHandle(STD_INPUT_HANDLE,fd1[0])) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }

 if(!SetStdHandle(STD_OUTPUT_HANDLE,fd2[1])) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }

 si.cb=sizeof(si);
 si.lpReserved=si.lpReserved2=0;
 si.cbReserved2=0;
 si.lpDesktop=0;
 si.lpTitle=0;
 si.dwFlags=STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
 si.wShowWindow=SW_HIDE;
 si.hStdInput=fd1[0];
 si.hStdOutput=fd2[1];
 si.hStdError=0;

 if(!DuplicateHandle(GetCurrentProcess(),fd1[1],GetCurrentProcess(),
     &dupWr,0,FALSE,DUPLICATE_SAME_ACCESS)) {
     pipe_close(fd1[0]);
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     pipe_close(fd2[1]);
     return 0;
 }
 pipe_close(fd1[1]);
 fd1[1]=dupWr;
          
 ret=CreateProcess(cmd,cmdline,0,0,TRUE,0,envblockp,dir,&si,&pi);
          
 pipe_close(fd1[0]); /*ommited for some reason from example*/
 pipe_close(fd2[1]);

 if(ret) {
     fd[0]=fd2[0];
     fd[1]=fd1[1];
 } else {
     pipe_close(fd1[1]);
     pipe_close(fd2[0]);
     warn_sys("CreateProcess %ld",GetLastError());
 }

 return ret;
}
#endif

#endif

