#if defined(UNIX)
#define OPENMODE_R O_RDONLY
#include <sys/types.h>
#include <sys/socket.h>
#elif defined(NT)
#include <windows.h>
#include <io.h>
#define OPENMODE_R (_O_RDONLY | _O_BINARY)
#endif

#include <stdio.h>
#include <errno.h>
#ifndef NT /* for srand() and time(0) and getpid() */
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#ifdef HAVE_DEV_RANDOM /* for open */
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "header.h"
#include "ps_sock.h"
#include "err_sys.h"
#include "stricmp.h"

#ifndef NT
#define closesocket close
#endif

struct PS_SOCKET_STRUCT {
     bool secure;
     bool ok;
     bool countSentBytes;  /*added by jim, for the HTTP log*/
     bool sentBytesCounter; /*ditto*/
     int maxSendSize;
     int maxRecvSize;
#ifdef USE_SSLEAY
     SSL * sslHandle;
#endif
     SOCKET pureSock;
};



struct PS_CTX_STRUCT {
    PS_SSL_CTX ssl_ctx;
    bool isServer;
    
    char serverkeyFilename[MAX_PATHNAME_SZ];
    char servercertFilename[MAX_PATHNAME_SZ];
    char CAsFilename[MAX_PATHNAME_SZ];
};



PS_SSL_CTX PS_CTX_getSSL_CTX(PS_CTX ps_ctx) {
    if(ps_ctx==NULL) return NULL;
    return ps_ctx->ssl_ctx;
}
char * PS_CTX_getServerKey(PS_CTX ps_ctx) {
    if(ps_ctx==NULL) return NULL;
    return ps_ctx->serverkeyFilename;
}
char * PS_CTX_getServerCert(PS_CTX ps_ctx) {
    if(ps_ctx==NULL) return NULL;
    return ps_ctx->servercertFilename;
}
char * PS_CTX_getCAs(PS_CTX ps_ctx) {
    if(ps_ctx==NULL) return NULL;
    return ps_ctx->CAsFilename;
}




static void PS_SSLeay_errors() {
#ifdef USE_SSLEAY
#ifdef PS_DEBUG
    int err;

    while ((err=ERR_get_error()))
        warn_sys("SSLeay %s",ERR_error_string(err,NULL));
#endif
#endif
}


#ifdef USE_SSLEAY


static RSA  *tmp_rsa_callback(SSL *s, int is_export, int keylength) {
	static RSA *rsa_tmp=NULL;

	if (rsa_tmp == NULL)
		{
		rsa_tmp=RSA_generate_key(keylength,RSA_F4,NULL,NULL);
		}
	return(rsa_tmp);
}

int verify_callback(int ok, X509_STORE_CTX *ctx)  {
    return(ok);
}

int verify_callback_true(int ok, X509_STORE_CTX *ctx)  {
    return 1;
}

#if !defined(NO_THREADS)
void ps_sock_set_locking_callback(locking_callback fct) {
 CRYPTO_set_locking_callback(fct);
}
#endif

#endif /*USE_SSLEAY*/

/* Ovrimos Rand will produce random numbers regardless of the platform we run. If it cannot use
 * system specific functions like /dev/urandom, etc, it will use the standard C functions.
 */
#define BUF_SIZE 256


static void _Ovrimos_Rand() {
#ifdef USE_SSLEAY
#ifdef NT
    char buff[BUF_SIZE];
    HCRYPTPROV hProv;
     
    if(!CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) {
        warn_sys("Error %x during CryptAcquireContext!\n", GetLastError());
        return;
    }

    if(!CryptGenRandom(hProv,  sizeof(buff), buff)) {
        warn_sys("Error %x during CryptGenRandom!\n", GetLastError());
        return;
    }
 
    RAND_seed(buff, sizeof(buff));

    if(!CryptReleaseContext(hProv, 0)) {
        warn_sys("Error %x during CryptReleaseContext!\n", GetLastError());
        return;
    }

#else /* no NT */

#ifdef HAVE_DEV_RANDOM
    char buff[BUF_SIZE];
    int fd;

    fd=open(NAME_OF_DEV_URANDOM, O_RDONLY);

    read(fd,buff,sizeof(buff));
    close(fd);

    RAND_seed(buff, sizeof(buff));

    return;

#else /* NO DEV_RANDOM */
/* Here we do a hack to get some random bits */
    uint32 buff[BUF_SIZE/sizeof(uint32)]; 
    int i;

    srand(getpid()*time(0));
    for (i=0;i<64;i++) buff[i]=rand();

    RAND_seed(buff, BUF_SIZE);
    
    return;
#endif
#endif /* NT */
#else /* NO SSLEAY */
    return;
#endif /* SSLEAY */
}


void PS_init() { /*automatically called by PS_newCtx()*/
#ifdef USE_SSLEAY
    static bool SSLeayInitialized = false;

    /*Initialize SSLeay essentials*/
    if(!SSLeayInitialized) {
        SSL_load_error_strings();
        SSL_library_init();
        SSLeayInitialized = true;

        _Ovrimos_Rand();

    }
#endif
}

PS_CTX PS_newCtx(bool isServer, const char *tlsCipherSuites, const char *own_cert_file, const char *own_key_file, const char *trusted_certs_file) {
    PS_SSL_CTX ssl_ctx = NULL;
    PS_CTX ctx = NULL;
#ifdef USE_SSLEAY
    int err;

    if(strempty(own_cert_file)) own_cert_file = NULL;
    if(strempty(own_key_file)) own_key_file = NULL;
    if(strempty(trusted_certs_file)) trusted_certs_file = NULL;


    PS_init();

    /*Create and initialize a new SSL context, used in further secure communications*/
    ssl_ctx = SSL_CTX_new(isServer?SSLv23_server_method():SSLv23_client_method());  



    if(ssl_ctx == NULL) {
          err_sys("Secure Server Context creation failed.");
    }

    SSL_CTX_set_options(ssl_ctx,SSL_OP_ALL);
    SSL_CTX_set_options(ssl_ctx,SSL_OP_NETSCAPE_DEMO_CIPHER_CHANGE_BUG);
    SSL_CTX_set_cipher_list(ssl_ctx,(char *)tlsCipherSuites);

    /*Load key & certificate*/
    if(own_cert_file && own_key_file) {

        err = SSL_CTX_use_certificate_file(ssl_ctx, own_cert_file, SSL_FILETYPE_PEM);
        if(err <= 0) {
             PS_SSLeay_errors();
             err_sys("Problem occured while reading certificate:%s",own_cert_file);
        }

        err = SSL_CTX_use_PrivateKey_file(ssl_ctx, own_key_file,  SSL_FILETYPE_PEM);
        if(err <= 0) {
             PS_SSLeay_errors();
             err_sys("Problem occured while attempting to read private key %s or entering its password.",own_key_file);
        }

    }


    SSL_CTX_set_tmp_rsa_callback(ssl_ctx,tmp_rsa_callback);

    /*Set trusted certificates file*/
    if(trusted_certs_file != NULL) {
      struct stat st;
      if(stat(trusted_certs_file,&st)<0) {
               err_sys("Trusted CA file cannot be accessed."); 
      } else {
          if(st.st_size>0) {
               if (!SSL_CTX_load_verify_locations(ssl_ctx, trusted_certs_file, NULL)) { 
                    PS_SSLeay_errors();
                    err_sys("SSL_CTX_load_verify_locations failure."); 
               }                    
               SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_PEER |
                    SSL_VERIFY_CLIENT_ONCE,
                    verify_callback);
          }
      }                           
    }


    ctx = (PS_CTX) malloc(sizeof(struct PS_CTX_STRUCT));
    ctx->ssl_ctx = ssl_ctx;
    ctx->isServer = isServer;
    if(own_cert_file != NULL) strcpy(ctx->servercertFilename, own_cert_file);
    if(own_key_file != NULL) strcpy(ctx->serverkeyFilename, own_key_file);
    if(trusted_certs_file != NULL)strcpy(ctx->CAsFilename, trusted_certs_file);
#endif
    return ctx;
}

PS_CTX PS_newSecureClientCtx(const char *sslroot, const char *tlsCipherSuits, bool authentication) {
    char own_cert_file[256];
    char own_key_file[256];
    char trusted_certs_file[256];

    if(sslroot == NULL) {
        sslroot = ".";
    }

    strcpy(own_cert_file, sslroot);
    strcat(own_cert_file, PS_CLIENTCERT);

    strcpy(own_key_file, sslroot);
    strcat(own_key_file, PS_CLIENTKEY);

    strcpy(trusted_certs_file, sslroot);
    strcat(trusted_certs_file, PS_CLIENTTRUSTEDCERTS);

    return PS_newCtx(false, tlsCipherSuits, own_cert_file, own_key_file, authentication?trusted_certs_file:NULL);
}


PS_CTX PS_newSecureServerCtx(const char *sslroot, const char *tlsCipherSuits, bool authentication) {
    char own_cert_file[256];
    char own_key_file[256];
    char trusted_certs_file[256];

    if(sslroot == NULL) {
        sslroot = ".";
    }


    strcpy(own_cert_file, sslroot);
    strcat(own_cert_file, PS_SERVERCERT);

    strcpy(own_key_file, sslroot);
    strcat(own_key_file, PS_SERVERKEY);

    strcpy(trusted_certs_file, sslroot);
    strcat(trusted_certs_file, PS_SERVERTRUSTEDCERTS);

    return PS_newCtx(true, tlsCipherSuits, own_cert_file, own_key_file, authentication?trusted_certs_file:NULL);
}

void PS_freeCtx(PS_CTX ctx) {
    if(ctx != NULL) {
#ifdef USE_SSLEAY    
     SSL_CTX_free(ctx->ssl_ctx);
#endif     
     free(ctx);
    }
}


PS_SOCKET PS_new() {
    PS_SOCKET ps_so=malloc(sizeof(struct PS_SOCKET_STRUCT));
    if(ps_so == NULL) {
        err_sys("Error while allocating memory for a PS_SOCKET");
    }
    ps_so->ok = false;
    ps_so->countSentBytes=false;
    ps_so->sentBytesCounter=0;
    return ps_so;
}
  
void PS_resetSentBytesCounter(PS_SOCKET ps_so) {
 ps_so->countSentBytes=true;
 ps_so->sentBytesCounter=0;
}

bool PS_getSentBytesCounter(PS_SOCKET ps_so, uint32 *p) {
 if(ps_so->countSentBytes) {
     *p=ps_so->sentBytesCounter;
 }
 return ps_so->countSentBytes;
}


SOCKET PS_getPureSocket(PS_SOCKET ps_so) {
    return ps_so->pureSock;
}


bool PS_handshake(PS_SOCKET pps_so, SOCKET so, PS_CTX ctx) {
    int sz;
#ifdef USE_SSLEAY
    int err;
#endif

    pps_so->pureSock = so;
    pps_so->secure = (ctx != NULL);
    pps_so->ok = false;
    sz = sizeof(pps_so->maxSendSize);
    if(getsockopt(so , SOL_SOCKET, SO_SNDBUF, (char*)&pps_so->maxSendSize, &sz)==-1) return false;
    sz = sizeof(pps_so->maxRecvSize);
    if(getsockopt(so , SOL_SOCKET, SO_RCVBUF, (char*)&pps_so->maxRecvSize, &sz)==-1) return false;
#ifndef USE_SSLEAY
    if(ctx != NULL)
        err_sys("internal error: SSL/TLS sockets are not supported in this version");
#else
    pps_so->sslHandle = NULL;


    if(ctx != NULL) {

        pps_so->sslHandle = SSL_new(ctx->ssl_ctx);

        if(pps_so->sslHandle  == NULL) {
            PS_SSLeay_errors();
            /*warn_sys("Secure Connection Not Established: Failed to create SSL handle.");*/
            return false;
        }

        SSL_clear(pps_so->sslHandle );

        if(!SSL_set_fd(pps_so->sslHandle , so)) {
            PS_SSLeay_errors();
            /*warn_sys("Secure Connection Not Established: Failed to associate socket with SSL handle.");*/
            return false;
        }

        if(ctx->isServer) {
            err = SSL_accept(pps_so->sslHandle);                  
        } else {
            err = SSL_connect(pps_so->sslHandle);                  
        }
        
        if (err <= 0) {         
            PS_SSLeay_errors();
            /*warn_sys("Secure Connection Not Established: Bad Handshake");*/
            return false;
        }

  }
#endif
  pps_so->ok = true;
  return true;
}


bool PS_isSecure(PS_SOCKET so) {
    return so->secure;
}

#ifdef USE_SSLEAY
static bool ps_X509NameExtract(X509_NAME *a, char *field, char *value) { /*Adaptod from SSLeay*/
    int len=127;
    char *buf = NULL;
    X509_NAME_ENTRY *ne;
    unsigned int i;
    int n,l2,num,j,type;
    const char * s;
    char *p;
    unsigned char *q;
    static char hex[17]="0123456789ABCDEF";
    int gs_doit[4];
    char tmp_buf[80];

    if (a == NULL) return false;

    value[0] = '\0';
    

    for (i=0; (int)i<sk_num((struct stack_st *)a->entries); i++)
        {
        ne=(X509_NAME_ENTRY *)sk_value((struct stack_st *)a->entries,i);
        n=OBJ_obj2nid(ne->object);
        if ((n == NID_undef) || ((s=OBJ_nid2sn(n)) == NULL))
            {
            i2t_ASN1_OBJECT(tmp_buf,sizeof(tmp_buf),ne->object);
            s=tmp_buf;
            }

        if(stricmp(s, field)==0)    buf = value; 
        else continue;
                                  

        type=ne->value->type;
        num=ne->value->length;
        q=ne->value->data;

        if ((type == V_ASN1_GENERALSTRING) && ((num%4) == 0))
            {
            gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=0;
            for (j=0; j<num; j++)
                if (q[j] != 0) gs_doit[j&3]=1;

            if (gs_doit[0]|gs_doit[1]|gs_doit[2])
                gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=1;
            else
                {
                gs_doit[0]=gs_doit[1]=gs_doit[2]=0;
                gs_doit[3]=1;
                }
            }
        else
            gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=1;

        for (l2=j=0; j<num; j++)
            {
            if (!gs_doit[j&3]) continue;
            l2++;
            if ((q[j] < ' ') || (q[j] > '~')) l2+=3;
            }

        if (l2 > len)
            {
            break;
            }
        else
            p= buf;

        q=ne->value->data;

        for (j=0; j<num; j++)
            {
            if (!gs_doit[j&3]) continue;
            n=q[j];
            if ((n < ' ') || (n > '~'))
                {
                *(p++)='\\';
                *(p++)='x';
                *(p++)=hex[(n>>4)&0x0f];
                *(p++)=hex[n&0x0f];
                }
            else
                *(p++)=n;
            }
        *p='\0';
        }
        
    p=buf;
    return true;
}
#endif

char *PS_getPeerCN(PS_SOCKET so, char *cn) { /*cn must have enought space for a CN*/
#ifdef USE_SSLEAY
    char *res;
    X509 *peer = SSL_get_peer_certificate(so->sslHandle);
	if (peer == NULL) return NULL;
    res = ps_X509NameExtract(X509_get_subject_name(peer), "CN", cn)? cn: NULL;
    X509_free(peer);
    return res;
#else
    return NULL;
#endif
}   

#if 0
static int foo(SSL *con)
	{
	char *buf=NULL;
	int ret=1;
	int i,j,k,blank,dot;
	struct stat st_buf;
	SSL *con;
	SSL_CIPHER *c;
	BIO *io,*ssl_bio,*sbio;
	long total_bytes;

	buf=Malloc(bufsize);
	if (buf == NULL) return(0);
	io=BIO_new(BIO_f_buffer());
	ssl_bio=BIO_new(BIO_f_ssl());
	if ((io == NULL) || (ssl_bio == NULL)) goto err;

#ifdef FIONBIO	
	if (s_nbio)
		{
		unsigned long sl=1;

		if (!s_quiet)
			BIO_printf(bio_err,"turning on non blocking io\n");
		if (BIO_socket_ioctl(s,FIONBIO,&sl) < 0)
			ERR_print_errors(bio_err);
		}
#endif

	/* lets make the output buffer a reasonable size */
	if (!BIO_set_write_buffer_size(io,bufsize)) goto err;

	if ((con=SSL_new(ctx)) == NULL) goto err;
	if(context) SSL_set_session_id_context(con, context,
					       strlen((char *)context));

	sbio=BIO_new_socket(s,BIO_NOCLOSE);
	if (s_nbio_test)
		{
		BIO *test;

		test=BIO_new(BIO_f_nbio_test());
		sbio=BIO_push(test,sbio);
		}
	SSL_set_bio(con,sbio,sbio);
	SSL_set_accept_state(con);

	/* SSL_set_fd(con,s); */
	BIO_set_ssl(ssl_bio,con,BIO_CLOSE);
	BIO_push(io,ssl_bio);
#ifdef CHARSET_EBCDIC
	io = BIO_push(BIO_new(BIO_f_ebcdic_filter()),io);
#endif

	if (s_debug)
		{
		con->debug=1;
		BIO_set_callback(SSL_get_rbio(con),bio_dump_cb);
		BIO_set_callback_arg(SSL_get_rbio(con),bio_s_out);
		}

	blank=0;
	for (;;)
		{
		if (hack)
			{
			i=SSL_accept(con);

			switch (SSL_get_error(con,i))
				{
			case SSL_ERROR_NONE:
				break;
			case SSL_ERROR_WANT_WRITE:
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_X509_LOOKUP:
				continue;
			case SSL_ERROR_SYSCALL:
			case SSL_ERROR_SSL:
			case SSL_ERROR_ZERO_RETURN:
				ret=1;
				goto err;
				/* break; */
				}

			SSL_renegotiate(con);
			SSL_write(con,NULL,0);
			}

		i=BIO_gets(io,buf,bufsize-1);
		if (i < 0) /* error */
			{
			if (!BIO_should_retry(io))
				{
				if (!s_quiet)
					ERR_print_errors(bio_err);
				goto err;
				}
			else
				{
				BIO_printf(bio_s_out,"read R BLOCK\n");
#ifndef MSDOS
				sleep(1);
#endif
				continue;
				}
			}
		else if (i == 0) /* end of input */
			{
			ret=1;
			goto end;
			}

		/* else we have data */
		if (	((www == 1) && (strncmp("GET ",buf,4) == 0)) ||
			((www == 2) && (strncmp("GET /stats ",buf,10) == 0)))
			{
			char *p;
			X509 *peer;
			STACK_OF(SSL_CIPHER) *sk;
			static char *space="                          ";

			BIO_puts(io,"HTTP/1.0 200 ok\r\nContent-type: text/html\r\n\r\n");
			BIO_puts(io,"<HTML><BODY BGCOLOR=\"#ffffff\">\n");
			BIO_puts(io,"<pre>\n");
/*			BIO_puts(io,SSLeay_version(SSLEAY_VERSION));*/
			BIO_puts(io,"\n");
			for (i=0; i<local_argc; i++)
				{
				BIO_puts(io,local_argv[i]);
				BIO_write(io," ",1);
				}
			BIO_puts(io,"\n");

			/* The following is evil and should not really
			 * be done */
			BIO_printf(io,"Ciphers supported in s_server binary\n");
			sk=SSL_get_ciphers(con);
			j=sk_SSL_CIPHER_num(sk);
			for (i=0; i<j; i++)
				{
				c=sk_SSL_CIPHER_value(sk,i);
				BIO_printf(io,"%-11s:%-25s",
					SSL_CIPHER_get_version(c),
					SSL_CIPHER_get_name(c));
				if ((((i+1)%2) == 0) && (i+1 != j))
					BIO_puts(io,"\n");
				}
			BIO_puts(io,"\n");
			p=SSL_get_shared_ciphers(con,buf,bufsize);
			if (p != NULL)
				{
				BIO_printf(io,"---\nCiphers common between both SSL end points:\n");
				j=i=0;
				while (*p)
					{
					if (*p == ':')
						{
						BIO_write(io,space,26-j);
						i++;
						j=0;
						BIO_write(io,((i%3)?" ":"\n"),1);
						}
					else
						{
						BIO_write(io,p,1);
						j++;
						}
					p++;
					}
				BIO_puts(io,"\n");
				}
			BIO_printf(io,((con->hit)
				?"---\nReused, "
				:"---\nNew, "));
			c=SSL_get_current_cipher(con);
			BIO_printf(io,"%s, Cipher is %s\n",
				SSL_CIPHER_get_version(c),
				SSL_CIPHER_get_name(c));
			SSL_SESSION_print(io,SSL_get_session(con));
			BIO_printf(io,"---\n");
			print_stats(io,SSL_get_SSL_CTX(con));
			BIO_printf(io,"---\n");
			peer=SSL_get_peer_certificate(con);
			if (peer != NULL)
				{
				BIO_printf(io,"Client certificate\n");
				X509_print(io,peer);
				PEM_write_bio_X509(io,peer);
				}
			else
				BIO_puts(io,"no client certificate available\n");
			BIO_puts(io,"</BODY></HTML>\r\n\r\n");
			break;
			}
		else if ((www == 2) && (strncmp("GET /",buf,5) == 0))
			{
			BIO *file;
			char *p,*e;
			static char *text="HTTP/1.0 200 ok\r\nContent-type: text/plain\r\n\r\n";

			/* skip the '/' */
			p= &(buf[5]);
			dot=0;
			for (e=p; *e != '\0'; e++)
				{
				if (e[0] == ' ') break;
				if (	(e[0] == '.') &&
					(strncmp(&(e[-1]),"/../",4) == 0))
					dot=1;
				}
			

			if (*e == '\0')
				{
				BIO_puts(io,text);
				BIO_printf(io,"'%s' is an invalid file name\r\n",p);
				break;
				}
			*e='\0';

			if (dot)
				{
				BIO_puts(io,text);
				BIO_printf(io,"'%s' contains '..' reference\r\n",p);
				break;
				}

			if (*p == '/')
				{
				BIO_puts(io,text);
				BIO_printf(io,"'%s' is an invalid path\r\n",p);
				break;
				}

			/* append if a directory lookup */
			if (e[-1] == '/')
				strcat(p,"index.html");

			/* if a directory, do the index thang */
			if (stat(p,&st_buf) < 0)
				{
				BIO_puts(io,text);
				BIO_printf(io,"Error accessing '%s'\r\n",p);
				ERR_print_errors(io);
				break;
				}
			if (S_ISDIR(st_buf.st_mode))
				{
				strcat(p,"/index.html");
				}

			if ((file=BIO_new_file(p,"r")) == NULL)
				{
				BIO_puts(io,text);
				BIO_printf(io,"Error opening '%s'\r\n",p);
				ERR_print_errors(io);
				break;
				}

			if (!s_quiet)
				BIO_printf(bio_err,"FILE:%s\n",p);

			i=strlen(p);
			if (	((i > 5) && (strcmp(&(p[i-5]),".html") == 0)) ||
				((i > 4) && (strcmp(&(p[i-4]),".php") == 0)) ||
				((i > 4) && (strcmp(&(p[i-4]),".htm") == 0)))
				BIO_puts(io,"HTTP/1.0 200 ok\r\nContent-type: text/html\r\n\r\n");
			else
				BIO_puts(io,"HTTP/1.0 200 ok\r\nContent-type: text/plain\r\n\r\n");
			/* send the file */
			total_bytes=0;
			for (;;)
				{
				i=BIO_read(file,buf,bufsize);
				if (i <= 0) break;

#ifdef RENEG
				total_bytes+=i;
				fprintf(stderr,"%d\n",i);
				if (total_bytes > 3*1024)
					{
					total_bytes=0;
					fprintf(stderr,"RENEGOTIATE\n");
					SSL_renegotiate(con);
					}
#endif

				for (j=0; j<i; )
					{
#ifdef RENEG
{ static count=0; if (++count == 13) { SSL_renegotiate(con); } }
#endif
					k=BIO_write(io,&(buf[j]),i-j);
					if (k <= 0)
						{
						if (!BIO_should_retry(io))
							goto write_error;
						else
							{
							BIO_printf(bio_s_out,"rwrite W BLOCK\n");
							}
						}
					else
						{
						j+=k;
						}
					}
				}
write_error:
			BIO_free(file);
			break;
			}
		}

	for (;;)
		{
		i=(int)BIO_flush(io);
		if (i <= 0)
			{
			if (!BIO_should_retry(io))
				break;
			}
		else
			break;
		}
end:
#if 1
	/* make sure we re-use sessions */
	SSL_set_shutdown(con,SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
#else
	/* This kills performance */
/*	SSL_shutdown(con); A shutdown gets sent in the
 *	BIO_free_all(io) procession */
#endif

err:

	if (ret >= 0)
		BIO_printf(bio_s_out,"ACCEPT\n");

	if (buf != NULL) Free(buf);
	if (io != NULL) BIO_free_all(io);
/*	if (ssl_bio != NULL) BIO_free(ssl_bio);*/
	return(ret);
	}

#endif

char *PS_getPeerEMail(PS_SOCKET so, char *email) { /*cn must have enought space for a CN*/
#ifdef USE_SSLEAY
    char dummy[1111];
    char *res;
    X509 *peer = SSL_get_peer_certificate(so->sslHandle);
	if (peer == NULL) return NULL;
    X509_NAME_oneline(X509_get_subject_name(peer), dummy, 1111);
    res = ps_X509NameExtract(X509_get_subject_name(peer), "EMail", email)? email: NULL;
    X509_free(peer);
//				X509_print(io,peer);
//				PEM_write_bio_X509(io,peer);

    return res;
#else
    return NULL;
#endif
}   


int PS_send_all(PS_SOCKET ps_so, const char *buf, int len) {
    int total_len=0;
    int maxSend = ps_so->maxSendSize;
    int n;

    if(!ps_so->ok) {
        warn_sys("PS_send_all:invalid socket");
        return -1;
    }

    if(!ps_so->secure) {
        do {
            n=send(ps_so->pureSock,buf,MIN(len,maxSend),0);
            if(n<0)  return n;
            if(n==0) break;
            total_len += n;
            buf += n;
            len -= n;
            ps_so->sentBytesCounter+=n;
        } while(len>0);
        return total_len;
    }
#ifdef USE_SSLEAY
    else {
        do {
            n=SSL_write(ps_so->sslHandle,(char *)buf,MIN(len,maxSend));
            if(n<0) {
                return n;
            }
            if(n==0) break;
            total_len +=n ;
            buf+=n;
            len-=n;
            ps_so->sentBytesCounter+=n;
        } while(len>0);
        return total_len;
    }
#else
     else {
          return -1; /* To shut up the warning */
     }
#endif
    
}


bool PS_recv_line(PS_SOCKET ps_so, char *buf, int len) {
    if(!ps_so->ok) {
       warn_sys("PS_recv_line:invalid socket");         
        return false;
    }

    if(!ps_so->secure) {
        int i;
        for(i=0; i<len; i++) {
            int n=recv(ps_so->pureSock,buf+i,1,0);
            if(n<=0) {
                return false;
            }
            if(buf[i]=='\n') {
                if(i>0 && buf[i-1]=='\r') {
                    buf[i-1]=0;
                }
                else {
                    buf[i]=0;
                }
                return true;
            }
        }
        return false;
    }
#ifdef USE_SSLEAY    
    else {
        int i;
        for(i=0; i<len; i++) {
            int err=SSL_read(ps_so->sslHandle,buf+i,1);
            if(err==-1)  {
                PS_SSLeay_errors();
                warn_sys("Error reading using SSL.");
                return false;
            }
            if(buf[i]=='\n') {
                if(i>0 && buf[i-1]=='\r') {
                     buf[i-1]=0;
                } else {
                    buf[i]=0;
                }
                return true;
            }
        }
        return false;
    }
#else
     else {
          return false; /* To shut up the warning */
     }
#endif
}


int PS_recv_all(PS_SOCKET ps_so, char *buf, int len) {
    int total_len = 0;
    int n;
    int maxRecv = ps_so->maxRecvSize;

    if(len == 0) return 0;
    if(!ps_so->secure) {
         do {
            n=recv(ps_so->pureSock, buf, MIN(len, maxRecv), 0);
            if(n<0)  return n;
            if(n==0) return total_len;
            total_len+=n;    
            buf+=n;
            len-=n;
        } while(len>0);
        return total_len;
    }
#ifdef USE_SSLEAY
    else {
         do {
            n=SSL_read(ps_so->sslHandle, buf, MIN(len, maxRecv));
            if(n<0)  return n;
            if(n==0) return total_len;
            total_len+=n;
            buf+=n;
            len-=n;
        } while(len>0);
        return total_len;
    }
#else
     else {
          return false; /* To shut up the warning */
     }
#endif
}


int PS_printf(PS_SOCKET s, const char *fmt, ...) {
    char buffer[50000];
    int len;
    va_list marker;

    va_start(marker,fmt);
    len=vsprintf(buffer,fmt,marker);
    if(PS_send_all(s,buffer,strlen(buffer))<=0) {
        len = -1;
    }
    va_end(marker);
    return len;
}



bool PS_send_file(PS_SOCKET so, const char *fname, char *buffer, int buffersize) {
    bool ret = true;
    int hd = open(fname,OPENMODE_R);
    if(hd!=-1) {
        int bytes;
        while(true) {
            bytes=read(hd,buffer,buffersize);
            if(bytes == -1) {ret = false; break;}
            if(bytes == 0)  {break;}
            if(PS_send_all(so,buffer,bytes) != bytes) {ret = false; break;}
        }
        close(hd);
        return ret; 
    } else { 
        
      perror( "Open failed " );
   
        return false;
    }
}

bool PS_send_file_range(PS_SOCKET so, const char *fname, char *buffer, uint32 buffersize, uint32 offset_s, uint32 offset_e) {
 bool ret = true;
 int hd = open(fname,OPENMODE_R);
 if(hd != -1) {
     int bytes;
     if(offset_s > 0) {
         if(lseek(hd,offset_s,SEEK_SET) == -1L) {close(hd); return false;} 
     }
     while(true) {
          uint32 howmany = offset_e - offset_s + 1;
          if(howmany>buffersize) {
               howmany=buffersize;
          }
          bytes=read(hd,buffer,howmany);
          if(bytes == -1) {ret = false; break;}
          if(bytes == 0)  {ret = offset_s == offset_e+1; break;}
          if(PS_send_all(so,buffer,bytes) != bytes) {ret = false; break;}
          offset_s += bytes;
     }
     close(hd);
     return ret; 
 } else {
    return false;
 }
}


int PS_closesocket(PS_SOCKET pps_so) {
    if(pps_so->ok) {
        pps_so->ok = false;
#ifdef USE_SSLEAY
        if(pps_so->secure) {
            SSL_free (pps_so->sslHandle);
        }
#endif
        return  closesocket(pps_so->pureSock);
    }
    return -1;
}

int PS_shutdown(PS_SOCKET pps_so, int mode) {
   return shutdown(pps_so->pureSock, mode);
}


void PS_release(PS_SOCKET ps_so) {
    PS_closesocket(ps_so);
    free(ps_so);
}

int PS_recv_all_timeout(PS_SOCKET ps_so, char *buf, int len, int secs, int millisecs) {
 int ret;
 fd_set readfs;
 struct timeval tm;

 FD_ZERO(&readfs);
 FD_SET(PS_getPureSocket(ps_so),&readfs);
 tm.tv_sec=secs;
 tm.tv_usec=millisecs;

 ret = select(1024,&readfs,0,0,&tm);

 if(ret==-1) {
     return -1;
 }
 if(FD_ISSET(PS_getPureSocket(ps_so),&readfs)) {
     return PS_recv_all(ps_so,buf,len);
 }
 return -1;
}

bool PS_authenticateServerAddress(PS_SOCKET ps_so, const char server[]) {
    char certCN[MAX_CN_SZ];

    if(!PS_isSecure(ps_so)) return true;

    if(!PS_getPeerCN(ps_so, certCN)) return false;

    if(certCN[0] != '*') {
        return stricmp(server, certCN) == 0;
    } else if(certCN[1] == '.') {
        return strend(server, certCN + 1);
    } else {
        return false;
    }
}

bool PS_matchesHost(const char host[], const char certCN[]) {
    if(certCN == NULL || host == NULL) return false;
    if(certCN[0] != '*') {
        return stricmp(host, certCN) == 0;
    } else if(certCN[1] == '.') {
        return strend(host, certCN + 1);
    } else {
        return false;
    }
}

